"""Contains the implementations for optimisation methods using BoTorch

Implements Bayesian optimisation methods using the BoTorch library.

"""

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from ax.plot.base import AxPlotConfig
from ax.plot.render import plot_config_to_html
from ax.service.ax_client import AxClient
from ax.service.utils.instantiation import ObjectiveProperties
from ax.utils.notebook.plotting import render
from ax.utils.report.render import render_report_elements
from plotly.subplots import make_subplots

from ..base import CalibrationWorkflowBase
from ..data_model import ParameterDataType, ParameterEstimateModel


class BoTorchOptimisation(CalibrationWorkflowBase):
	"""The BoTorchOptimisation optimisation method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []

		parameter_spec = self.specification.parameter_spec.parameters
		parameters = []
		objectives = {}

		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			data_type = spec.data_type
			self.data_types.append(data_type)
			if data_type == ParameterDataType.DISCRETE:
				value_type = "int"
			else:
				value_type = "float"

			bounds = self.get_parameter_bounds(spec)

			parameter = dict(
				name=parameter_name,
				type="range",
				bounds=list(bounds),
				value_type=value_type,
			)
			parameters.append(parameter)

		n_out = self.specification.n_out
		self.output_labels = self.specification.output_labels
		directions = self.specification.directions
		if self.output_labels is None:
			self.output_labels = [f"objective_{i}" for i in range(n_out)]

		for i, output_label in enumerate(self.output_labels):
			direction = directions[i]
			if direction == "minimize":
				minimize = True
			else:
				minimize = False
			objectives[output_label] = ObjectiveProperties(minimize=minimize)

		n_init = self.specification.n_init
		random_seed = self.specification.random_seed
		n_jobs = self.specification.n_jobs
		use_saasbo = self.specification.use_saasbo
		choose_generation_strategy_kwargs = dict(
			max_initialization_trials=n_init,
			random_seed=random_seed,
			max_parallelism_cap=n_jobs,
			use_saasbo=use_saasbo,
		)

		self.ax_client = AxClient(verbose_logging=False)
		self.ax_client.create_experiment(
			name=self.specification.experiment_name,
			parameters=parameters,
			objectives=objectives,
			choose_generation_strategy_kwargs=choose_generation_strategy_kwargs,
			overwrite_existing_experiment=True,
		)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""

		objective_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: dict[str, float]) -> np.ndarray:
			X = list(X.values())
			X = [X]

			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				objective_kwargs,
				True,
			)
			Y = Y[0]

			results = {}
			for i, output in enumerate(self.output_labels):  # type: ignore[arg-type]
				results[output] = (Y[i], 0.0)

			return results

		n_iterations = self.specification.n_iterations
		for _ in range(n_iterations):
			parameterization, trial_index = self.ax_client.get_next_trial()
			self.ax_client.complete_trial(
				trial_index=trial_index, raw_data=target_function(parameterization)
			)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		def plot_config(config: AxPlotConfig, title: str) -> None:
			fig = plot_config_to_html(config)
			outfile = self.join(
				outdir,  # type: ignore[arg-type]
				f"{time_now}-{task}-{experiment_name}-{title}.html",
			)

			if outdir is not None:
				with open(outfile, "w") as f:
					f.write(
						render_report_elements(
							title,
							html_elements=[fig],
							header=False,
						)
					)
				self.append_artifact(outfile)
			else:
				render(config)

		config = self.ax_client.get_optimization_trace()
		plot_config(config, "optimization-trace")
		config = self.ax_client.get_feature_importances()
		plot_config(config, "feature-importances")

		trials = []
		for trial in self.ax_client.experiment.trials.values():
			if isinstance(trial.arm, list):
				for arm in trial.arm:
					parameters = arm.parameters
					parameters["arm_name"] = arm.name
					trials.append(parameters)
			else:
				parameters = trial.arm.parameters
				parameters = {f"param_{k}": parameters[k] for k in parameters}
				parameters["arm_name"] = trial.arm.name
				trials.append(parameters)

		trials_df = pd.DataFrame(trials).set_index("arm_name")
		objective_df = self.ax_client.experiment.fetch_data().df.set_index("arm_name")
		trials_df = (
			trials_df.join(objective_df)
			.reset_index()
			.sort_values("mean", ascending=True)
		)

		if outdir is not None:
			self.to_csv(trials_df, "objective")

		parameter_names = [col for col in trials_df if col.startswith("param_")]
		fig = make_subplots(
			rows=1, cols=len(parameter_names), subplot_titles=parameter_names
		)
		for i, parameter_name in enumerate(parameter_names):
			fig.add_trace(
				go.Scatter(
					x=trials_df[parameter_name], y=trials_df["mean"], mode="markers"
				),
				row=1,
				col=i + 1,
			)

		fig.update_layout(yaxis_title="Score", showlegend=False)
		if outdir is not None:
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-plot-slice.png"
			)
			self.append_artifact(outfile)
			fig.write_image(outfile)
		else:
			fig.show()

		fig = go.Figure(
			data=go.Scatter(
				x=trials_df["trial_index"], y=trials_df["mean"], mode="markers"
			)
		)
		fig.update_layout(xaxis_title="Trial", yaxis_title="Score", showlegend=False)
		if outdir is not None:
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-trial-history.png"
			)
			fig.write_image(outfile)
			self.append_artifact(outfile)
		else:
			fig.show()

		values = [value for value in trials_df.columns if value.startswith("mean")]

		trials_df_best = trials_df.sort_values(values).head(1)
		for col in trials_df_best.columns:
			if not col.startswith("param_"):
				continue
			name = col.replace("param_", "")
			estimate = trials_df_best[col].item()
			parameter_estimate = ParameterEstimateModel(name=name, estimate=estimate)
			self.add_parameter_estimate(parameter_estimate)



"""Contains the implementations for the quadrature methods

Implements the supported quadrature methods.

"""

from collections.abc import Callable

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .emukit_wrapper import EmukitQuadrature

TASK = "quadrature"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	emukit=EmukitQuadrature
)


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for quadrature.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary
			of calibration implementations for quadrature.
	"""
	return IMPLEMENTATIONS


class QuadratureMethodModel(CalibrationModel):
	"""The quadrature method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	kernel: str | None = Field(
		description="The Kernel embeddings for Bayesian quadrature",
		default="QuadratureRBFLebesgueMeasure",
	)
	measure: str | None = Field(
		description="The Integration measures", default="LebesgueMeasure"
	)


class QuadratureMethod(CalibrationMethodBase):
	"""The quadrature method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: QuadratureMethodModel,
		engine: str = "emukit",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""QuadratureMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (QuadratureMethodModel): The calibration
				specification.
		    engine (str, optional): The Quadrature backend.
				Defaults to "emukit".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains the implementations for the likelihood-free methods

Implements the supported likelihood-free methods.

"""

import importlib
from collections.abc import Callable

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel

TASK = "likelihood_free"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict()

if importlib.util.find_spec("elfi") is not None:
	from ..experimental.likelihood_free.elfi_wrapper import ELFILikelihoodFree

	IMPLEMENTATIONS["elfi"] = ELFILikelihoodFree


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for likelihood-free.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary of
			calibration implementations for likelihood-free.
	"""
	return IMPLEMENTATIONS


class LikelihoodFreeMethodModel(CalibrationModel):
	"""The likelihood-free method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	epsilon: float = Field(
		description="The dissimilarity threshold between observed and simulated data",
		default=0,
	)
	acq_noise_var: float = Field(
		description="Noise added to Lower Confidence Bound Selection Criterion",
		default=0,
	)
	sampler: str = Field(
		description="The sampling algorithm to perform inference", default="metropolis"
	)


class LikelihoodFreeMethod(CalibrationMethodBase):
	"""The likelihood-free method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: LikelihoodFreeMethodModel,
		engine: str = "elfi",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""LikelihoodFreeMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (LikelihoodFreeMethodModel): The
				calibration specification.
		    engine (str, optional): The likelihood-free backend.
				Defaults to "elfi".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains the OpenTurns base class

The defined base class for performing OpenTurns.

"""

from collections.abc import Callable

import numpy as np
import openturns as ot

from .calibration_base import CalibrationWorkflowBase


class OpenTurnsBase(CalibrationWorkflowBase):
	"""The OpenTurns base class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []
		self.bounds: tuple[list[float], list[float]] = ([], [])
		parameters = []

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			bounds = self.get_parameter_bounds(spec)
			lower_bound, upper_bound = bounds
			lower_bounds, upper_bounds = self.bounds
			lower_bounds.append(lower_bound)
			upper_bounds.append(upper_bound)

			data_type = spec.data_type
			self.data_types.append(data_type)

			distribution_name = (
				spec.distribution_name.replace("_", " ").title().replace(" ", "")
			)
			distribution_args = spec.distribution_args
			if distribution_args is None:
				distribution_args = []

			distribution_kwargs = spec.distribution_kwargs
			if distribution_kwargs is None:
				distribution_kwargs = {}

			dist_instance = getattr(ot, distribution_name)
			parameter = dist_instance(*distribution_args, **distribution_kwargs)
			parameters.append(parameter)

		distribution_collection = ot.DistributionCollection(parameters)
		self.parameters = ot.JointDistribution(distribution_collection)
		self.parameters.setDescription(self.names)

	def get_ot_func_wrapper(self, func_sample: Callable) -> Callable:
		"""Get an OpenTurns wrapper for Python functions.

		Args:
		    func_sample (Callable): The function to wrap.

		Returns:
		    Callable: The wrapped function.
		"""
		n_dim = self.parameters.getDimension()
		n_out = self.specification.n_out

		return ot.PythonFunction(n_dim, n_out, func_sample=func_sample)

	def replicate_samples(self, X: ot.Sample) -> ot.Sample:
		"""Replicate the sampled parameters.

		Args:
			X (ot.Sample): The sampled parameters.

		Returns:
			ot.Sample: The replicated samples.
		"""
		n_replicates = self.specification.n_replicates
		if n_replicates > 1:
			X = X.asDataFrame().values
			X = np.repeat(X, n_replicates, axis=0)
			self.rng.shuffle(X)
			X = ot.Sample(X)
		return X

	def sample_parameters(self, n_samples: int) -> np.ndarray:
		"""Get new parameter samples.

		Args:
		    n_samples (int): The number of samples.

		Returns:
		    np.ndarray: The parameter samples.
		"""
		if not hasattr(self, "experiment"):
			self.experiment = ot.LHSExperiment(self.parameters, n_samples)

		self.experiment.setSize(n_samples)
		X = self.experiment.generate()
		X = self.replicate_samples(X)

		return X

	def get_X_Y(
		self, n_samples: int, target_function: Callable
	) -> tuple[np.ndarray, np.ndarray]:
		"""Get the X and Y matrices.

		Args:
		    n_samples (int): The number of samples to take
		        from the random design.
		    target_function (Callable):
		        The simulation function.

		Returns:
		    tuple[np.ndarray, np.ndarray]: The X and Y matrices.
		"""
		X = self.specification.X
		if X is None:
			X = self.sample_parameters(n_samples)

		Y = self.specification.Y
		if Y is None:
			ot_func_wrapper = self.get_ot_func_wrapper(target_function)
			Y = ot_func_wrapper(X)

		return X, Y



"""Contains the implementations for Bayesian quadrature methods using Emukit

Implements the supported Bayesian quadrature methods using the Emukit library.

"""

import emukit.quadrature.kernels
import emukit.quadrature.measures
import numpy as np
from emukit.model_wrappers.gpy_quadrature_wrappers import BaseGaussianProcessGPy, RBFGPy
from emukit.quadrature.loop import VanillaBayesianQuadratureLoop
from emukit.quadrature.methods import VanillaBayesianQuadrature
from GPy.models import GPRegression
from matplotlib import pyplot as plt

from ..base import EmukitBase


class EmukitQuadrature(EmukitBase):
	"""The Emukit Bayesian quadrature method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		quadrature_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: np.ndarray) -> np.ndarray:
			return self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				quadrature_kwargs,
			)

		n_init = self.specification.n_init
		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		X, Y = self.get_X_Y(n_init, target_function)
		gp = GPRegression(X, Y, **method_kwargs)
		emukit_rbf = RBFGPy(gp.kern)

		measure_name = self.specification.measure
		measure_func = getattr(emukit.quadrature.measures, measure_name)
		measure = measure_func.from_bounds(bounds=self.bounds)

		kernel_name = self.specification.kernel
		kernel_func = getattr(emukit.quadrature.kernels, kernel_name)
		kernel = kernel_func(emukit_rbf, measure)
		emulator = BaseGaussianProcessGPy(kern=kernel, gpy_model=gp)

		quadrature = VanillaBayesianQuadrature(base_gp=emulator, X=X, Y=Y)
		quadrature_loop = VanillaBayesianQuadratureLoop(model=quadrature)
		n_iterations = self.specification.n_iterations
		quadrature_loop.run_loop(target_function, stopping_condition=n_iterations)

		self.emulator = emulator
		self.quadrature_loop = quadrature_loop

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		integral_mean, integral_variance = self.quadrature_loop.model.integrate()
		fig, ax = plt.subplots(figsize=self.specification.figsize)
		ax.set_title("Integral density")
		self.rng = self.get_default_rng(self.specification.random_seed)
		integral_samples = self.rng.normal(
			integral_mean, integral_variance, size=self.specification.n_samples
		)
		ax.hist(
			integral_samples,
			alpha=0.5,
		)
		ax.legend()
		self.present_fig(
			fig, outdir, time_now, task, experiment_name, "integral-density"
		)



from .design import get_full_factorial_design
from .discrepancy import (
	BrayCurtisDistance,
	CanberraDistance,
	ChebyshevDistance,
	CorrelationDistance,
	CosineDistance,
	DistanceMetricBase,
	EnergyDistance,
	GaussianLogLikelihood,
	JensenShannonDistance,
	KlDivergence,
	L1Norm,
	L2Norm,
	MeanAbsoluteError,
	MeanAbsolutePercentageError,
	MeanPinballLoss,
	MeanSquaredError,
	MedianAbsoluteError,
	MinkowskiDistance,
	MultivariateNormalLogLikelihood,
	PoissonLogLikelihood,
	RootMeanSquaredError,
	StudentsTLogLikelihood,
	WassersteinDistance,
	get_distance_metric_func,
)

__all__ = [
	get_full_factorial_design,
	BrayCurtisDistance,
	CanberraDistance,
	ChebyshevDistance,
	CorrelationDistance,
	CosineDistance,
	DistanceMetricBase,
	GaussianLogLikelihood,
	get_distance_metric_func,
	KlDivergence,
	JensenShannonDistance,
	L1Norm,
	L2Norm,
	WassersteinDistance,
	EnergyDistance,
	MinkowskiDistance,
	MeanSquaredError,
	MeanAbsoluteError,
	MultivariateNormalLogLikelihood,
	PoissonLogLikelihood,
	RootMeanSquaredError,
	StudentsTLogLikelihood,
	MeanPinballLoss,
	MeanAbsolutePercentageError,
	MedianAbsoluteError,
]



"""Contains the implementations for optimisation methods using
OpenTurns

Implements the supported optimisation methods using
the OpenTurns library.

"""

import numpy as np
import openturns as ot
import openturns.viewer as viewer
import pandas as pd

from ..base import OpenTurnsBase
from ..data_model import ParameterEstimateModel
from ..estimators import KrigingEstimator


class OpenTurnsOptimisation(OpenTurnsBase):
	"""The OpenTurns optimisation method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		method = self.specification.method

		supported_methods = list(ot.Pagmo.GetAlgorithmNames())
		supported_methods.append("kriging")

		if method not in supported_methods:
			raise ValueError(
				f"Unsupported optimisation algorithm: {method}.",
				f"Supported optimisation algorithms are {', '.join(supported_methods)}",
			)

		optimisation_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: np.ndarray) -> np.ndarray:
			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				optimisation_kwargs,
			)
			if len(Y.shape) == 1:
				Y = np.expand_dims(Y, axis=1)
			return Y

		ot_func_wrapper = self.get_ot_func_wrapper(target_function)
		self.problem = ot.OptimizationProblem(ot_func_wrapper)
		lower_bounds, upper_bounds = self.bounds
		bounds = ot.Interval(lower_bounds, upper_bounds)
		self.problem.setBounds(bounds)

		n_init = self.specification.n_init
		n_iterations = self.specification.n_iterations

		if method == "kriging":
			X, Y = self.get_X_Y(n_init, target_function)

			method_kwargs = self.specification.method_kwargs
			if method_kwargs is None:
				method_kwargs = {}

			method_kwargs["covariance_scale"] = method_kwargs.get(
				"covariance_scale", 1.0
			)
			method_kwargs["covariance_amplitude"] = method_kwargs.get(
				"covariance_amplitude", 1.0
			)
			method_kwargs["parameters"] = self.parameters
			method_kwargs["n_out"] = self.specification.n_out
			estimator = KrigingEstimator(**method_kwargs)

			estimator.fit(X, Y)
			self.emulator = estimator

			self.study = ot.EfficientGlobalOptimization(self.problem, estimator.result)
			self.study.setMaximumCallsNumber(n_iterations)
			self.study.run()
		else:
			X = self.specification.X
			if X is None:
				X = self.sample_parameters(n_init)

			self.study = ot.Pagmo(self.problem, method, X)
			self.study.setMaximumIterationNumber(n_iterations)
			self.study.run()

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		method = self.specification.method

		result = self.study.getResult()
		graph = result.drawOptimalValueHistory()
		view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
		if outdir is not None:
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-plot-optimization-history.png",
			)
			self.append_artifact(outfile)
			view.save(outfile)

		graph = result.drawErrorHistory()
		view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
		if outdir is not None:
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-plot-error-history.png"
			)
			self.append_artifact(outfile)
			view.save(outfile)

		parameters = result.getInputSample()
		objective_results = result.getOutputSample()

		trial_rows = []
		for i, parameter_set in enumerate(parameters):
			trial_row = dict(algorithm=method, number=i)

			objective_result = objective_results[i]
			for j, objective_result in enumerate(objective_result):
				trial_row[f"value_{j}"] = objective_result

			for j, name in enumerate(self.names):
				trial_row[f"param_{name}"] = parameter_set[j]

			trial_rows.append(trial_row)
		trials_df: pd.DataFrame = pd.DataFrame(trial_rows)

		values = [value for value in trials_df.columns if value.startswith("value")]
		trials_df_best = trials_df.sort_values(values).head(1)
		for col in trials_df_best.columns:
			if not col.startswith("param_"):
				continue
			name = col.replace("param_", "")
			estimate = trials_df_best[col].item()
			parameter_estimate = ParameterEstimateModel(name=name, estimate=estimate)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(trials_df, "trials")



"""Contains utility functions for calibration

This module defines various utility functions
for the calibration of simulations.

"""

import os.path as osp
import uuid
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd

from ..data_model import ParameterDataType


def get_datetime_now() -> str:
	"""Get the current datetime for now.

	Returns:
	    str: The current datetime.
	"""
	return datetime.today().strftime("%Y-%m-%d-%H-%M-%S")


def get_simulation_uuid() -> str:
	"""Get a new simulation uuid.

	Returns:
	    str: The simulation uuid.
	"""
	simulation_uuid = str(uuid.uuid4())
	return simulation_uuid


def get_examples_outdir() -> str:  # pragma: no cover
	"""Get the output directory for calibration examples.

	Returns:
	    str: The output directory.
	"""
	return osp.join("examples", "outdir")


def create_file_path(file_path: str) -> str:
	"""Create file path if it does not exist.

	Args:
		file_path (str): The file path to create.

	Returns:
		str: The created file path.
	"""
	path = Path(file_path)
	if not path.is_dir():
		path.mkdir(parents=True, exist_ok=True)
	return file_path


def calibration_func_wrapper(
	X: np.ndarray,
	workflow: "CalibrationWorkflowBase",  # type: ignore[name-defined] # noqa: F821
	observed_data: pd.DataFrame | np.ndarray,
	parameter_names: list[str],
	data_types: list[ParameterDataType],
	calibration_kwargs: dict,
	wrap_values: bool = False,
) -> np.ndarray:
	"""Wrapper function for the calibration function.

	Args:
		X (np.ndarray): The parameter set matrix.
		workflow (CalibrationWorkflowBase): The calibration workflow.
		observed_data (pd.DataFrame | np.ndarray): The observed data.
		parameter_names (list[str]): The list of simulation parameter names.
		data_types (list[ParameterDataType]): The data types for each parameter.
		calibration_kwargs (dict): Arguments to supply to the calibration function.
		wrap_values (bool): Whether to wrap scalar values with a list.
			Defaults to False.

	Returns:
		np.ndarray: The simulation output data.
	"""
	import numpy as np

	parameters = []
	for theta in X:
		parameter_set = {}
		for i, parameter_value in enumerate(theta):
			parameter_name = parameter_names[i]
			data_type = data_types[i]
			if data_type == ParameterDataType.CONTINUOUS:
				parameter_set[parameter_name] = parameter_value
			else:
				parameter_set[parameter_name] = int(parameter_value)
		parameters.append(parameter_set)

	simulation_ids = [get_simulation_uuid() for _ in range(len(parameters))]

	if workflow.specification.batched:
		results = workflow.call_calibration_func(
			parameters, simulation_ids, observed_data, **calibration_kwargs
		)
	else:
		results = []
		for i, parameter in enumerate(parameters):
			simulation_id = simulation_ids[i]
			result = workflow.call_calibration_func(
				parameter,
				simulation_id,
				observed_data,
				**calibration_kwargs,
			)
			if wrap_values and not isinstance(result, list):
				result = [result]
			results.append(result)
	results = np.array(results)
	return results


class EarlyStopper:
	"""Early stopping for training."""

	def __init__(self, patience: int = 1, min_delta: float = 0):
		"""EarlyStopper constructor.

		Args:
		    patience (int, optional): The number of iterations
		           before performing early stopping. Defaults to 1.
		    min_delta (float, optional): The minimum difference
		       for the validation loss. Defaults to 0.
		"""
		self.patience = patience
		self.min_delta = min_delta
		self.counter: int = 0
		self.min_validation_loss = float("inf")

	def early_stop(self, validation_loss: float) -> bool:
		"""Perform early stopping.

		Args:
		    validation_loss (float): The training validation loss.

		Returns:
		    bool: Whether to perform early stopping.
		"""
		if validation_loss < self.min_validation_loss:
			self.min_validation_loss = validation_loss
			self.counter = 0
		elif validation_loss >= (self.min_validation_loss + self.min_delta):
			self.counter += 1
			if self.counter >= self.patience:
				return True
		return False


def extend_X(X: np.ndarray, Y_rows: int) -> np.ndarray:
	"""Extend the number of rows for X with a dummy index column.

	Args:
		X (np.ndarray): The input matrix.
		Y_rows (int) The number of rows for the simulation outputs.

	Returns:
		np.ndarray: The extended input matrix with a dummy column.
	"""
	design_list = []
	for i in range(X.shape[0]):
		for j in range(Y_rows):
			row = np.append(X[i], j)
			design_list.append(row)
	X = np.array(design_list)
	return X



"""Contains the implementations for state estimation using
Ensemble Kalman Filter

Implements the supported state estimation using
the Ensemble Kalman Filter.

"""

# mypy: disable-error-code="index, arg-type"

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from ...base import HistoryMatchingBase

# See https://github.com/ajwdewit/pcse_notebooks/blob/master/08a%20Data%20assimilation%20with%20the%20EnKF.ipynb


class EKFStateEstimation(HistoryMatchingBase):
	"""The Ensemble Kalman Filter state estimation method class."""

	def get_ensemble(self) -> dict[str, dict]:
		"""Return the ensemble.

		Returns:
		    dict[str, list]: The ensemble.
		"""
		return self.ensemble

	def initialise_ensemble(self) -> None:
		"""Initialise the state of all ensemble members."""
		ensemble_size = self.specification.n_samples
		observed_data = self.specification.observed_data
		outputs = self.specification.output_labels
		state_estimation_kwargs = self.get_calibration_func_kwargs()

		self.t = -1
		self.observed_indx = 0
		self.perform_update = False
		self.ensemble = {}

		for i in range(ensemble_size):
			simulation_id = self.get_simulation_uuid()
			parameters: dict[str, float] = {}
			self.ensemble[simulation_id] = dict(
				parameters=parameters, model=None, result={}
			)

			for output in outputs:
				self.ensemble[simulation_id]["result"][output] = []  # type: ignore[assignment]

			for k in self.parameters:
				parameters[k] = self.parameters[k][i]

			self.call_calibration_func(
				parameters, simulation_id, observed_data, **state_estimation_kwargs
			)

		self.t = 0

	def update_step(self) -> None:
		"""Perform the update/analysis step."""
		outputs = self.specification.output_labels
		ensemble_size = self.specification.n_samples
		observed_data = self.specification.observed_data
		stds = self.specification.stds

		collected_states = []
		for simulation_id in self.ensemble:
			state = {}
			for output in outputs:
				state_variable = self.ensemble[simulation_id]["result"][output][-1]
				state[output] = state_variable
			collected_states.append(state)

		df_A = pd.DataFrame(collected_states)
		A = np.matrix(df_A).T
		if np.isnan(A).any():
			return
		P_e = np.matrix(df_A.cov())

		perturbed_observations = []
		for output in outputs:
			observed_value = observed_data[output].iloc[self.observed_indx]
			std = abs(observed_value * stds[output])
			sample = self.rng.normal(observed_value, std, (ensemble_size))
			perturbed_observations.append(sample)

		df_perturbed_observations = pd.DataFrame(perturbed_observations).T
		df_perturbed_observations.columns = outputs
		D = np.matrix(df_perturbed_observations).T
		R_e = np.matrix(df_perturbed_observations.cov())

		H = np.identity(len(outputs))
		K1 = P_e * (H.T)
		K2 = (H * P_e) * H.T
		K = K1 * (np.linalg.pinv(K2 + R_e))
		Aa = A + K * (D - (H * A))
		df_Aa = pd.DataFrame(Aa.T, columns=outputs)

		replace_state_variables = self.specification.replace_state_variables
		for i, simulation_id in enumerate(self.ensemble):
			for output in outputs:
				updated_variable = df_Aa[output].iloc[i]
				if replace_state_variables:
					self.ensemble[simulation_id]["result"][output][-1] = (
						updated_variable
					)
				else:
					self.ensemble[simulation_id]["result"][output].append(
						updated_variable
					)

		self.perform_update = False
		self.observed_indx += 1

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		pass_calibration_workflow = self.specification.pass_calibration_workflow
		if pass_calibration_workflow is None:
			self.specification.pass_calibration_workflow = True

		observed_data = self.specification.observed_data
		state_estimation_kwargs = self.get_calibration_func_kwargs()

		self.initialise_ensemble()

		n_iterations = self.specification.n_iterations
		for _ in range(n_iterations):
			for simulation_id in self.ensemble:
				parameters = self.ensemble[simulation_id]["parameters"]
				self.call_calibration_func(
					parameters, simulation_id, observed_data, **state_estimation_kwargs
				)

			if self.perform_update:
				self.update_step()

			self.t += 1

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		observed_data = self.specification.observed_data

		outputs = self.specification.output_labels
		results: dict[str, list] = dict(simulation_id=[], index=[])
		for output in outputs:
			results[output] = []

		for simulation_id in self.ensemble:
			result: list[float] | float = []
			for output in outputs:
				result = self.ensemble[simulation_id]["result"][output]
				results[output].extend(result)

			result_len = len(result)
			simulation_ids = [simulation_id for _ in range(result_len)]
			results["simulation_id"].extend(simulation_ids)
			results["index"].extend(np.arange(0, result_len, 1).tolist())
		results_df = pd.DataFrame(results)

		for output in outputs:
			fig, axes = plt.subplots(
				nrows=2, figsize=self.specification.figsize, sharex=False, sharey=False
			)

			observed_data[output].plot(ax=axes[0])
			for _, df in results_df.groupby("simulation_id"):
				df.plot(x="index", y=output, ax=axes[1], legend=False)

			self.present_fig(
				fig, outdir, time_now, task, experiment_name, f"ensemble-{output}"
			)

		def q1(x: float) -> float:
			return x.quantile(0.025)

		def q2(x: float) -> float:
			return x.quantile(0.5)

		def q3(x: float) -> float:
			return x.quantile(0.975)

		quantile_list = [q1, q2, q3]
		for output in outputs:
			agg = {output: quantile_list}
			quantile_df = (
				results_df.groupby("index", as_index=False)
				.agg(agg)
				.droplevel(axis=1, level=0)
				.reset_index()
			)

			fig, axes = plt.subplots(
				nrows=2, figsize=self.specification.figsize, sharex=False, sharey=False
			)
			observed_data[output].plot(ax=axes[0])
			for quantile in quantile_list:
				quantile_df[quantile.__name__].plot(ax=axes[1])

			self.present_fig(
				fig,
				outdir,
				time_now,
				task,
				experiment_name,
				f"ensemble-quantile-{output}",
			)

		if outdir is None:
			return

		results_df = results_df.drop(columns="index")
		outfile = self.join(outdir, f"{time_now}-{task}-{experiment_name}_ensemble.csv")
		self.append_artifact(outfile)
		results_df.to_csv(outfile, index=False)



"""Contains the implementations for the experimental design methods

Implements the supported experimental design methods.

"""

from collections.abc import Callable

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .emukit_wrapper import EmukitExperimentalDesign

TASK = "experimental_design"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	emukit=EmukitExperimentalDesign
)


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for experimental design.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary
			of calibration implementations for experimental design.
	"""
	return IMPLEMENTATIONS


class ExperimentalDesignMethodModel(CalibrationModel):
	"""The experimental design method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""


class ExperimentalDesignMethod(CalibrationMethodBase):
	"""The experimental design method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: ExperimentalDesignMethodModel,
		engine: str = "emukit",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""ExperimentalDesignMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (ExperimentalDesignMethodModel): The calibration
				specification.
		    engine (str, optional): The experimental design backend.
				Defaults to "emukit".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains the implementations for active learning methods using scikit-activeml

Implements the supported active learning methods using the scikit-activeml library.

"""

import numpy as np
import pandas as pd
from emukit.core.initial_designs import RandomDesign
from matplotlib import pyplot as plt
from skactiveml.pool import (
	ExpectedModelChangeMaximization,
	ExpectedModelVarianceReduction,
	GreedySamplingTarget,
	GreedySamplingX,
	RegressionTreeBasedAL,
)
from skactiveml.regressor import NICKernelRegressor, SklearnRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.tree import DecisionTreeRegressor

from ..base import EmukitBase


class SkActiveMLActiveLearning(EmukitBase):
	"""The scikit-activeml active learning method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		active_learning_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: np.ndarray) -> np.ndarray:
			return self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				active_learning_kwargs,
			)

		n_init = self.specification.n_init
		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		X, Y_true = self.get_X_Y(n_init, target_function)
		self.Y_shape = 1
		if len(Y_true.shape) > 1:
			self.Y_shape = Y_true.shape[1]
			if self.Y_shape > 1:
				X = self.extend_X(X, self.Y_shape)
				Y_true = Y_true.flatten()

		Y = np.full_like(Y_true, np.nan)
		surrogate_name = self.specification.method
		surrogates = dict(
			nick=NICKernelRegressor,
			gp=GaussianProcessRegressor,
			rf=RandomForestRegressor,
			dt=DecisionTreeRegressor,
		)
		surrogate_class = surrogates.get(surrogate_name, None)
		if surrogate_class is None:
			raise ValueError(
				f"Unsupported surrogate class: {surrogate_name}.",
				f"Supported surrogate classes are {', '.join(surrogates)}",
			)

		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		emulator = surrogate_class(**method_kwargs)
		if surrogate_name != "nick":
			emulator = SklearnRegressor(emulator)

		query_name = self.specification.query_strategy
		query_stategies = dict(
			greedy_sampling_x=GreedySamplingX,
			greedy_sampling_target=GreedySamplingTarget,
			regression_tree_based_al=RegressionTreeBasedAL,
			# kl_divergence_maximization=KLDivergenceMaximization,
			expected_model_change_maximization=ExpectedModelChangeMaximization,
			expected_model_variance_reduction=ExpectedModelVarianceReduction,
		)
		query_class = query_stategies.get(query_name, None)
		if query_class is None:
			raise ValueError(
				f"Unsupported query strategy: {query_name}.",
				f"Supported query strategies are {', '.join(query_stategies)}",
			)
		query_strategy = query_class(random_state=self.specification.random_seed)

		n_iterations = self.specification.n_iterations
		for _ in range(n_iterations):
			query_idx = query_strategy.query(X=X, y=Y, reg=emulator, fit_reg=True)
			Y[query_idx] = Y_true[query_idx]

		emulator.fit(X, Y)
		self.emulator = emulator
		self.query_strategy = query_strategy
		self.Y_true = Y_true

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		design = RandomDesign(self.parameter_space)
		n_samples = self.specification.n_samples
		X_sample = design.get_samples(n_samples)
		if self.Y_shape > 1:
			X_sample = self.extend_X(X_sample, self.Y_shape)
		predicted = self.emulator.predict(X_sample)

		names = self.names.copy()
		output_labels = self.specification.output_labels
		if output_labels is None:
			output_labels = ["output"]
		output_label = output_labels[0]
		if X_sample.shape[1] > len(names):
			names.append("_dummy_index")
		df = pd.DataFrame(X_sample, columns=names)
		df[f"emulated-{output_label}"] = predicted

		if self.specification.use_shap and outdir is not None:
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-param-importances.png",
			)
			self.calculate_shap_importances(
				X_sample, self.emulator, names, self.specification.test_size, outfile
			)

		fig, axes = plt.subplots(
			nrows=len(self.names), figsize=self.specification.figsize
		)
		for i, name in enumerate(self.names):
			df.plot.scatter(name, f"emulated-{output_label}", ax=axes[i])
		self.present_fig(
			fig, outdir, time_now, task, experiment_name, f"emulated-{output_label}"
		)

		if outdir is None:
			return

		self.to_csv(df, f"emulated-{output_label}")



"""Contains the Scikit-Learn wrappers for OpenTurns

The defined Scikit-Learn wrappers for the OpenTurns library.

"""

import copy

import numpy as np
import openturns as ot
from sklearn.base import BaseEstimator, MultiOutputMixin, RegressorMixin
from sklearn.metrics import r2_score
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y


class FunctionalChaosEstimator(MultiOutputMixin, RegressorMixin, BaseEstimator):
	def __init__(self, parameters: ot.JointDistribution, total_degree: int = 2):
		"""FunctionalChaosEstimator constructor.

		Args:
		    parameters (dict): The distribution of parameters.
		    total_degree (int, optional): The total polynomial degree.
		        Defaults to 2.
		"""
		self.parameters = parameters

		input_dim = parameters.getDimension()
		columns = [
			ot.StandardDistributionPolynomialFactory(self.parameters.getMarginal(i))
			for i in range(input_dim)
		]
		enumerate_func = ot.LinearEnumerateFunction(input_dim)
		product_basic = ot.OrthogonalProductPolynomialFactory(columns, enumerate_func)
		candidate_basis = enumerate_func.getBasisSizeFromTotalDegree(total_degree)
		self.adaptive_strategy = ot.FixedStrategy(product_basic, candidate_basis)
		self.projection_strategy = ot.LeastSquaresStrategy()

		self.emulator = None

	def fit(
		self, X: np.ndarray, y: np.ndarray | None = None
	) -> "FunctionalChaosEstimator":
		"""Fit the estimator.

		Args:
		    X (np.ndarray): The simulation inputs.
		    y (np.ndarray | None, optional): The simulation outputs.
		        Defaults to None.

		Returns:
		    FunctionalChaosEstimator: The estimator.
		"""
		X = np.array(X)
		y = np.array(y)
		if len(y.shape) == 1:
			y = np.expand_dims(y, axis=1)

		X, y = check_X_y(
			X,
			y,
			multi_output=True,
			y_numeric=True,
			ensure_2d=True,
			dtype="numeric",
		)
		self.is_fitted_ = True
		self.n_features_in_ = X.shape[1]

		self.algo = ot.FunctionalChaosAlgorithm(
			X, y, self.parameters, self.adaptive_strategy, self.projection_strategy
		)
		self.algo.run()
		self.result = self.algo.getResult()
		self.emulator = self.result.getMetaModel()

		return self

	def predict(self, X: np.ndarray) -> np.ndarray:
		"""Make a prediction.

		Args:
		    X (np.ndarray): The simulation inputs.

		Returns:
		    np.ndarray: The model predictions.
		"""
		check_is_fitted(self, "is_fitted_")
		X = check_array(X, ensure_2d=True, dtype="numeric")

		y_pred = self.emulator(X)
		return np.array(y_pred)

	def score(
		self, X: np.ndarray, y: np.ndarray, sample_weight: np.ndarray | None = None
	) -> float:
		"""Assess the estimator.

		Args:
		    X (np.ndarray): The simulation inputs.
		    y (np.ndarray): The simulation outputs.
		    sample_weight (np.ndarray | None, optional): Weighting factor
		        for the samples. Defaults to None.

		Returns:
		    float: The assessment score.
		"""
		y_pred = self.emulator(X)
		return r2_score(y_pred, y, sample_weight=sample_weight)


class KrigingEstimator(MultiOutputMixin, RegressorMixin, BaseEstimator):
	def __init__(
		self,
		parameters: ot.JointDistribution,
		basis: str = "constant",
		covariance: str = "SquaredExponential",
		covariance_scale: float = 1.0,
		covariance_amplitude: float = 1.0,
		n_out: int = 1,
	):
		"""KrigingEstimator constructor.

		Args:
		    parameters (dict): The distribution of parameters.
		    basis (str, optional): The basis function.
		        Defaults to "constant".
		    covariance (str, optional): The covariance function.
		        Defaults to "SquaredExponential".
		    covariance_scale (float, optional): Scale coefficient.
		        Defaults to 1.0.
		    covariance_amplitude (float, optional): Amplitude of the process.
		        Defaults to 1.0.
		    n_out (int, optional): The number of outputs.
		"""
		self.parameters = parameters
		input_dim = parameters.getDimension()

		supported_basis = dict(
			constant=ot.ConstantBasisFactory,
			linear=ot.LinearBasisFactory,
			quadratic=ot.QuadraticBasisFactory,
		)
		self.basis = supported_basis.get(basis, ot.ConstantBasisFactory)(
			input_dim
		).build()

		if n_out > 1:
			self.basis = ot.Basis(
				[
					ot.AggregatedFunction([self.basis.build(k)] * n_out)
					for k in range(self.basis.getSize())
				]
			)

		self.covariance = getattr(ot, covariance)(
			[covariance_scale] * input_dim, [covariance_amplitude]
		)

		if n_out > 1:
			self.covariance = ot.TensorizedCovarianceModel(
				[copy.deepcopy(self.covariance) for _ in range(n_out)]
			)

		self.emulator = None

	def fit(self, X: np.ndarray, y: np.ndarray | None = None) -> "KrigingEstimator":
		"""Fit the estimator.

		Args:
		    X (np.ndarray): The simulation inputs.
		    y (np.ndarray | None, optional): The simulation outputs.
		        Defaults to None.

		Returns:
		    KrigingEstimator: The estimator.
		"""
		X = np.array(X)
		y = np.array(y)
		if len(y.shape) == 1:
			y = np.expand_dims(y, axis=1)

		X, y = check_X_y(
			X,
			y,
			multi_output=True,
			y_numeric=True,
			ensure_2d=True,
			dtype="numeric",
		)
		self.is_fitted_ = True
		self.n_features_in_ = X.shape[1]

		self.algo = ot.KrigingAlgorithm(X, y, self.covariance, self.basis)
		self.algo.run()
		self.result = self.algo.getResult()
		self.emulator = self.result.getMetaModel()

		return self

	def predict(
		self,
		X: np.ndarray,
	) -> np.ndarray | tuple:
		"""Make a prediction.

		Args:
		    X (np.ndarray): The simulation inputs.

		Returns:
		    np.ndarray | tuple: The model predictions.
		"""
		check_is_fitted(self, "is_fitted_")
		X = check_array(X, ensure_2d=True, dtype="numeric")
		y_pred = self.emulator(X)
		return np.array(y_pred)

	def score(
		self, X: np.ndarray, y: np.ndarray, sample_weight: np.ndarray | None = None
	) -> float:
		"""Assess the estimator.

		Args:
		    X (np.ndarray): The simulation inputs.
		    y (np.ndarray): The simulation outputs.
		    sample_weight (np.ndarray | None, optional): Weighting factor
		        for the samples. Defaults to None.

		Returns:
		    float: The assessment score.
		"""
		y_pred = self.emulator(X)
		return r2_score(y_pred, y, sample_weight=sample_weight)



from .simulation_data_models import (
	CalibrationModel,
	DistributionModel,
	ParameterDataType,
	ParameterEstimateModel,
	ParameterEstimatesModel,
	ParameterModel,
	ParameterSpecification,
)

__all__ = [
	ParameterModel,
	DistributionModel,
	ParameterDataType,
	CalibrationModel,
	ParameterSpecification,
	ParameterEstimateModel,
	ParameterEstimatesModel,
]



"""Contains the implementations for Bayesian calibration methods using
emcee

Implements the supported Bayesian calibration methods using
the emcee library.

"""

from multiprocessing import Pool

import corner
import emcee
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from ..base import CalibrationWorkflowBase
from ..data_model import ParameterEstimateModel


def target_function(X: np.ndarray, self: CalibrationWorkflowBase) -> np.ndarray:
	bayesian_calibration_kwargs = self.get_calibration_func_kwargs()
	lower_bounds, upper_bounds = self.bounds
	for i, x in enumerate(X):
		if x < lower_bounds[i] or x > upper_bounds[i]:
			return -np.inf

	X = np.array([X])
	Y = self.calibration_func_wrapper(
		X,
		self,
		self.specification.observed_data,
		self.names,
		self.data_types,
		bayesian_calibration_kwargs,
	)
	if len(Y.shape) == 1:
		Y = np.expand_dims(Y, axis=1)

	return Y


class EmceeBayesianCalibration(CalibrationWorkflowBase):
	"""The emcee Bayesian calibration method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []
		self.bounds: tuple[list[float], list[float]] = ([], [])

		parameter_spec = self.specification.parameter_spec.parameters
		n_walkers = self.specification.n_samples
		self.rng = self.get_default_rng(self.specification.random_seed)

		self.parameters = []
		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			data_type = spec.data_type
			self.data_types.append(data_type)

			bounds = spec.distribution_bounds
			lower_bound, upper_bound = bounds
			lower_bounds, upper_bounds = self.bounds
			lower_bounds.append(lower_bound)
			upper_bounds.append(upper_bound)

			distribution_name = spec.distribution_name.replace(" ", "_").lower()

			distribution_args = spec.distribution_args
			if distribution_args is None:
				distribution_args = []

			distribution_kwargs = spec.distribution_kwargs
			if distribution_kwargs is None:
				distribution_kwargs = {}

			dist_instance = getattr(self.rng, distribution_name)
			samples = dist_instance(
				*distribution_args, **distribution_kwargs, size=n_walkers
			)
			self.parameters.append(samples)

		self.parameters = np.array(self.parameters).T

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		n_walkers, n_dim = self.parameters.shape

		supported_moves = dict(
			RedBlueMove=emcee.moves.RedBlueMove,
			StretchMove=emcee.moves.StretchMove,
			WalkMove=emcee.moves.WalkMove,
			KDEMove=emcee.moves.KDEMove,
			DEMove=emcee.moves.DEMove,
			DESnookerMove=emcee.moves.DESnookerMove,
			MHMove=emcee.moves.MHMove,
			GaussianMove=emcee.moves.GaussianMove,
		)

		moves_spec = self.specification.moves
		if moves_spec is None:
			moves = None
		else:
			moves = []
			for name, weight in moves_spec.items():
				move = supported_moves.get(name)
				if move is None:
					raise ValueError(
						f"Unsupported move: {name}.",
						f"Supported moves are {', '.join(supported_moves)}",
					)
				moves.append((move, weight))

		n_jobs = self.specification.n_jobs
		n_iterations = self.specification.n_iterations
		verbose = self.specification.verbose

		if n_jobs > 1:
			with Pool(n_jobs) as pool:
				self.sampler = emcee.EnsembleSampler(
					n_walkers,
					n_dim,
					target_function,
					args=[self],
					moves=moves,
					pool=pool,
					vectorize=False,
				)
				self.sampler.run_mcmc(
					self.parameters, n_iterations, tune=True, progress=verbose
				)
		else:
			self.sampler = emcee.EnsembleSampler(
				n_walkers,
				n_dim,
				target_function,
				args=[self],
				moves=moves,
				vectorize=False,
			)
			self.sampler.run_mcmc(
				self.parameters, n_iterations, tune=True, progress=verbose
			)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		samples = self.sampler.get_chain()

		labels = self.names
		n_dim = len(labels)
		fig, axes = plt.subplots(n_dim, figsize=self.specification.figsize, sharex=True)
		for i in range(n_dim):
			ax = axes[i]
			ax.plot(samples[:, :, i], "k", alpha=0.3)
			ax.set_xlim(0, len(samples))
			ax.set_ylabel(labels[i])
			ax.yaxis.set_label_coords(-0.1, 0.5)

		axes[-1].set_xlabel("Step number")
		self.present_fig(fig, outdir, time_now, task, experiment_name, "trace")

		flat_samples = self.sampler.get_chain(flat=True)
		fig = corner.corner(flat_samples, labels=labels)
		self.present_fig(fig, outdir, time_now, task, experiment_name, "plot_surface")

		trace_df = pd.DataFrame(flat_samples, columns=labels)
		fig, axes = plt.subplots(nrows=n_dim, figsize=self.specification.figsize)

		for i, parameter_name in enumerate(labels):
			axes[i].set_title(parameter_name)
			axes[i].hist(trace_df[parameter_name], label=parameter_name)

		self.present_fig(fig, outdir, time_now, task, experiment_name, "plot_slice")

		for name in trace_df:
			estimate = trace_df[name].mean()
			uncertainty = trace_df[name].std()

			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(trace_df, "trace")



from .implementation import (
	ReliabilityAnalysisMethod,
	ReliabilityAnalysisMethodModel,
	get_implementations,
)

_all__ = [
	ReliabilityAnalysisMethod,
	ReliabilityAnalysisMethodModel,
	get_implementations,
]



"""Contains the implementations for the simulation-based inference methods

Implements the supported simulation-based inference methods.

"""

import importlib
from collections.abc import Callable

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel

TASK = "simulation_based_inference"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict()

if importlib.util.find_spec("lampe") is not None:
	from .lampe_wrapper import LAMPESimulationBasedInference

	IMPLEMENTATIONS["lampe"] = LAMPESimulationBasedInference

if importlib.util.find_spec("sbi") is not None:
	from .sbi_wrapper import SBISimulationBasedInference

	IMPLEMENTATIONS["sbi"] = SBISimulationBasedInference


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for simulation-based inference.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary of
			calibration implementations for simulation-based inference.
	"""
	return IMPLEMENTATIONS


class SimulationBasedInferenceMethodModel(CalibrationModel):
	"""The simulation-based inference method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	num_simulations: int = Field(
		description="The number of simulations to run", default=25
	)


class SimulationBasedInferenceMethod(CalibrationMethodBase):
	"""The simulation-based inference method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: SimulationBasedInferenceMethodModel,
		engine: str = "sbi",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""SimulationBasedInferenceMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (SimulationBasedInferenceMethodModel): The
				calibration specification.
		    engine (str, optional): The simulation-based inference
				backend. Defaults to "sbi".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



from .implementation import (
	UncertaintyAnalysisMethod,
	UncertaintyAnalysisMethodModel,
	get_implementations,
)

_all__ = [
	UncertaintyAnalysisMethod,
	UncertaintyAnalysisMethodModel,
	get_implementations,
]



from .implementation import (
	SimulationBasedInferenceMethod,
	SimulationBasedInferenceMethodModel,
	get_implementations,
)

_all__ = [
	SimulationBasedInferenceMethod,
	SimulationBasedInferenceMethodModel,
	get_implementations,
]



from .implementation import (
	ApproximateBayesianComputationMethod,
	ApproximateBayesianComputationMethodModel,
	get_implementations,
)

_all__ = [
	ApproximateBayesianComputationMethod,
	ApproximateBayesianComputationMethodModel,
	get_implementations,
]



from .anharmonic_oscillator import AnharmonicOscillator
from .lorenz_95 import Lorenz95
from .lotka_volterra import LotkaVolterraModel
from .sir_ode import SirOdesModel

__all__ = [LotkaVolterraModel, SirOdesModel, Lorenz95, AnharmonicOscillator]



from .utilities import (
	EarlyStopper,
	calibration_func_wrapper,
	create_file_path,
	extend_X,
	get_datetime_now,
	get_examples_outdir,
	get_simulation_uuid,
)

__all__ = [
	calibration_func_wrapper,
	create_file_path,
	extend_X,
	get_datetime_now,
	get_simulation_uuid,
	get_examples_outdir,
	EarlyStopper,
]



"""Contains Pydantic data models for the simulation

Several Pydantic data models are defined for various
simulation calibration procedures.

"""

from enum import Enum
from typing import Any

import numpy as np
import pandas as pd
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field


class BaseModel(PydanticBaseModel):
	"""Base Pydantic data model.

	Args:
	    PydanticBaseModel (PydanticBaseModel): The Pydantic Base model class.
	"""

	class Config:
		arbitrary_types_allowed = True
		protected_namespaces = ()


class ParameterDataType(Enum):
	DISCRETE: str = "discrete"
	CONTINUOUS: str = "continuous"
	CATEGORICAL: str = "categorical"


class ParameterModel(BaseModel):
	"""The simulation parameter data model.

	Args:
	    BaseModel (BaseModel): The Pydantic Base model class.
	"""

	name: str = Field(description="The parameter name")
	parameter_values: list[float] | None = Field(
		description="The list of parameter values.", default=None
	)
	parameter_tags: dict[str, str] | None = Field(
		description="A collection of metadata tags.", default=None
	)
	data_type: ParameterDataType = Field(
		description="The parameter data type", default=ParameterDataType.CONTINUOUS
	)


class DistributionModel(ParameterModel):
	"""The probability distribution data model.

	Args:
	    ParameterModel (ParameterModel): The simulation parameter data model.
	"""

	distribution_name: str = Field(
		description="The distribution name", default="uniform"
	)
	distribution_bounds: list | None = Field(
		description="The distribution bounds", default=None
	)
	distribution_args: list | None = Field(
		description="The distribution positional arguments", default=None
	)
	distribution_kwargs: dict[str, Any] | None = Field(
		description="The distribution named arguments", default=None
	)


class ParameterSpecification(BaseModel):
	"""The collection of parameters.

	Args:
	    BaseModel (BaseModel): The Pydantic Base model class.
	"""

	parameters: list[DistributionModel] | None = Field(
		description="The parameter specification list", default=None
	)


class CalibrationModel(BaseModel):
	"""The calibration data model.

	Args:
	    BaseModel (BaseModel): The Pydantic Base model class.
	"""

	parameter_spec: ParameterSpecification | None = Field(
		description="The parameter specification", default=None
	)
	experiment_name: str | None = Field(
		description="The modelling experiment name", default="default"
	)
	outdir: str | None = Field(
		description="The output directory for modelling results", default=None
	)
	method: str = Field(description="The calibration method or algorithm", default="")
	method_kwargs: dict[str, Any] | None = Field(
		description="The calibration method named arguments", default=None
	)
	calibration_func_kwargs: dict[str, Any] | None = Field(
		description="The calibration function named arguments", default=None
	)
	pass_calibration_workflow: bool | str | None = Field(
		description="Pass the calibration workflow into the calibration function",
		default=None,
	)
	analyze_kwargs: dict[str, Any] | None = Field(
		description="The analyze step named arguments", default=None
	)
	X: np.ndarray | pd.DataFrame | list | None = Field(
		description="The simulation input data", default=None
	)
	Y: np.ndarray | pd.DataFrame | list | None = Field(
		description="The simulation output data", default=None
	)
	test_size: float | int = Field(
		description="The size of the testing dataset.", default=0.0
	)
	n_out: int = Field(description="Number of simulation outputs", default=1)
	observed_data: np.ndarray | pd.DataFrame | float | None = Field(
		description="The empirical or observed data", default=None
	)
	n_samples: int = Field(description="The number of samples to take", default=1)
	n_chains: int = Field(description="The number of Markov chains", default=1)
	n_init: int = Field(description="The number of initial samples or steps", default=1)
	n_iterations: int = Field(
		description="The number of iterations for sequential calibrators", default=1
	)
	lr: float = Field(
		description="The learning rate of the model",
		default=0.01,
	)
	random_seed: int | None = Field(
		description="The random seed for replicability", default=None
	)
	use_shap: bool = Field(
		description="Whether to use SHAP for feature explanations", default=False
	)
	n_replicates: int = Field(
		description="The number of replicate simulations to run", default=1
	)
	n_jobs: int = Field(description="The number of jobs to run in parallel", default=1)
	storage: str | None = Field(
		description="The storage backend to persist the calibration results",
		default=None,
	)
	walltime: int = Field(description="The maximum calibration walltime", default=1)
	output_labels: list[str] | None = Field(
		description="The list of simulation output names", default=None
	)
	groups: list[str] | None = Field(
		description="The list of parameter groups", default=None
	)
	parallel_backend: str = Field(
		description="The backend engine to run parallel jobs", default=""
	)
	batched: bool = Field(description="Whether to batch the simulations", default=False)
	verbose: bool = Field(
		description="Whether to print calibration messages", default=False
	)
	figsize: tuple[int, int] = Field(
		description="The figure size for visualisations", default=(12, 12)
	)


class ParameterEstimateModel(BaseModel):
	"""The simulation parameter estimate data model.

	Args:
	    BaseModel (BaseModel): The Pydantic Base model class.
	"""

	name: str = Field(description="The parameter name.")
	estimate: float = Field(
		description="The estimated parameter value; a point estimate"
	)
	uncertainty: float | None = Field(
		description="The uncertainty of the parameter estimate", default=None
	)


class ParameterEstimatesModel(BaseModel):
	"""The simulation parameter estimates data model.

	Args:
	    BaseModel (BaseModel): The Pydantic Base model class.
	"""

	estimates: list[ParameterEstimateModel] | None = Field(
		description="The set of parameter estimates", default=None
	)



"""Contains the implementations for the sensitivity analysis methods

Implements the supported sensitivity analysis methods.

"""

from collections.abc import Callable

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .openturns_wrapper import OpenTurnsSensitivityAnalysis
from .salib_wrapper import SALibSensitivityAnalysis

TASK = "sensitivity_analysis"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	salib=SALibSensitivityAnalysis,
	openturns=OpenTurnsSensitivityAnalysis,
)


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for sensitivity analysis.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary of
			calibration implementations for sensitivity analysis.
	"""
	return IMPLEMENTATIONS


class SensitivityAnalysisMethodModel(CalibrationModel):
	"""The sensitivity analysis method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration
			base model class.
	"""

	order: int = Field(
		description="The order for the polynomial chaos expansion", default=2
	)


class SensitivityAnalysisMethod(CalibrationMethodBase):
	"""The sensitivity analysis method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: SensitivityAnalysisMethodModel,
		engine: str = "salib",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""SensitivityAnalysisMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (SensitivityAnalysisMethodModel): The calibration
				specification.
		    engine (str, optional): The sensitivity analysis backend.
				Defaults to "salib".
			implementation (CalibrationWorkflowBase | None): The calibration
				workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains the implementations for surrogate modelling methods using Scikit-Learn

Implements the supported surrogate modelling methods using the Scikit-Learn library.

"""

import numpy as np
import pandas as pd
import sklearn.ensemble as ensemble
import sklearn.gaussian_process as gp
import sklearn.kernel_ridge as kernel_ridge
import sklearn.linear_model as lm
import sklearn.neighbors as neighbors
import sklearn.svm as svm
from matplotlib import pyplot as plt

from ..base import SurrogateBase
from ..estimators import EmukitEstimator, FunctionalChaosEstimator, KrigingEstimator


class SklearnSurrogateModel(SurrogateBase):
	"""The Scikit-Learn surrogate modelling method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		surrogate_kwargs = self.get_calibration_func_kwargs()
		n_samples = self.specification.n_samples

		X = self.specification.X
		if X is None:
			X = self.sample_parameters(n_samples)

		Y = self.specification.Y
		if Y is None:
			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				surrogate_kwargs,
			)

		emulator_name = self.specification.method
		emulators = dict(
			gp=gp.GaussianProcessRegressor,
			emukit_gp=EmukitEstimator,
			openturns_functional_chaos=FunctionalChaosEstimator,
			openturns_kriging=KrigingEstimator,
			rf=ensemble.RandomForestRegressor,
			gb=ensemble.GradientBoostingRegressor,
			lr=lm.LinearRegression,
			elastic=lm.ElasticNet,
			ridge=lm.Ridge,
			knn=neighbors.KNeighborsRegressor,
			kr=kernel_ridge.KernelRidge,
			linear_svm=svm.LinearSVR,
			nu_svm=svm.NuSVR,
		)
		emulator_class = emulators.get(emulator_name, None)
		if emulator_class is None:
			raise ValueError(
				f"Unsupported emulator: {emulator_name}.",
				f"Supported emulators are {', '.join(emulators)}",
			)

		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		self.Y_shape = Y.shape
		if (
			self.specification.flatten_Y
			and len(self.Y_shape) > 1
			and self.specification.X is None
		):
			X = self.extend_X(X, self.Y_shape[1])
			Y = Y.flatten()

		self.emulator = emulator_class(**method_kwargs)
		self.emulator.fit(X, Y)

		self.X = X
		self.Y = Y

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		output_labels = self.specification.output_labels
		if output_labels is None:
			output_labels = ["output"]
		output_label = output_labels[0]

		names = self.names.copy()
		if self.X.shape[1] > len(names):
			names.append("_dummy_index")
		df = pd.DataFrame(self.X, columns=names)

		n_samples = self.specification.n_samples
		X_sample = self.sample_parameters(n_samples)
		if self.specification.flatten_Y and len(self.Y_shape) > 1:
			X_sample = self.extend_X(X_sample, self.Y_shape[1])
		Y_sample = self.emulator.predict(X_sample)

		if self.specification.use_shap and outdir is not None:
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-param-importances.png",
			)
			self.calculate_shap_importances(
				X_sample, self.emulator, names, self.specification.test_size, outfile
			)

		if len(self.Y_shape) == 1:
			df[f"simulated_{output_label}"] = self.Y
			fig, axes = plt.subplots(
				nrows=len(self.names), figsize=self.specification.figsize
			)
			if not isinstance(axes, np.ndarray):
				axes = [axes]

			for i, parameter_name in enumerate(self.names):
				df.plot.scatter(
					parameter_name,
					f"simulated_{output_label}",
					ax=axes[i],
					title=f"simulated_{output_label} against {parameter_name}",
				)
			self.present_fig(fig, outdir, time_now, task, experiment_name, "plot-slice")

			fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)
			df = pd.DataFrame(
				{
					"index": np.arange(0, self.Y.shape[0], 1),
					"simulated": self.Y,
					"emulated": Y_sample,
				}
			)
			df.plot.scatter(
				"index", "simulated", ax=axes[0], title=f"Simulated {output_label}"
			)
			df.plot.scatter(
				"index", "emulated", ax=axes[1], title=f"Emulated {output_label}"
			)
			self.present_fig(
				fig, outdir, time_now, task, experiment_name, f"emulated-{output_label}"
			)
		else:
			if self.specification.flatten_Y:
				df[f"simulated_{output_label}"] = self.Y
				fig, axes = plt.subplots(
					nrows=len(self.names), figsize=self.specification.figsize
				)
				if not isinstance(axes, np.ndarray):
					axes = [axes]

				for i, parameter_name in enumerate(self.names):
					df.plot.scatter(
						parameter_name,
						f"simulated_{output_label}",
						ax=axes[i],
						title=f"simulated_{output_label} against {parameter_name}",
					)
				self.present_fig(
					fig, outdir, time_now, task, experiment_name, "plot-slice"
				)

				Y_sample = Y_sample.reshape(self.Y_shape)
				Y = self.Y.reshape(self.Y_shape)
				index = np.arange(0, Y_sample.shape[1], 1)

				fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)

				for i in range(Y.shape[0]):
					axes[0].plot(index, Y[i])
				axes[0].set_title(f"Simulated {output_label}")

				for i in range(Y_sample.shape[0]):
					axes[1].plot(index, Y_sample[i])
				axes[1].set_title(f"Emulated {output_label}")
				self.present_fig(
					fig,
					outdir,
					time_now,
					task,
					experiment_name,
					f"emulated-{output_label}",
				)
			else:
				output_labels = self.specification.output_labels
				if len(output_labels) != self.Y_shape[-1]:  # type: ignore[arg-type]
					output_labels = [f"Output_{y}" for y in range(self.Y_shape[-1])]

				fig, axes = plt.subplots(
					nrows=len(self.names) * len(output_labels),  # type: ignore[arg-type]
					ncols=2,
					figsize=self.specification.figsize,
				)

				row_index = 0
				for x_index, parameter_name in enumerate(self.names):
					for y_index, output_label in enumerate(output_labels):  # type: ignore[arg-type]
						axes[row_index, 0].scatter(
							self.X[:, x_index], self.Y[:, y_index]
						)
						axes[row_index, 0].set_xlabel(parameter_name)
						axes[row_index, 0].set_ylabel(output_label)
						axes[row_index, 0].set_title(f"Simulated {output_label}")

						axes[row_index, 1].scatter(
							X_sample[:, x_index], Y_sample[:, y_index]
						)
						axes[row_index, 1].set_xlabel(parameter_name)
						axes[row_index, 1].set_ylabel(output_label)
						axes[row_index, 1].set_title(f"Emulated {output_label}")

						row_index += 1

				self.present_fig(
					fig, outdir, time_now, task, experiment_name, "plot-slice"
				)



"""Contains the implementations for Approximate Bayesian Computation methods using
PyMC

Implements the supported Approximate Bayesian Computation methods using
the PyMC library.

"""

from collections.abc import Callable

import arviz as az
import numpy as np
import pymc as pm
from matplotlib import pyplot as plt

from ..base import CalibrationWorkflowBase
from ..data_model import ParameterEstimateModel


class PyMCApproximateBayesianComputation(CalibrationWorkflowBase):
	"""The PyMC Approximate Bayesian Computation method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		priors = []
		parameter_spec = self.specification.parameter_spec.parameters
		with pm.Model() as self.model:
			for spec in parameter_spec:
				parameter_name = spec.name
				self.names.append(parameter_name)

				distribution_name = (
					spec.distribution_name.replace("_", " ").title().replace(" ", "")
				)

				distribution_class = getattr(pm, distribution_name)
				distribution_args = spec.distribution_args
				if distribution_args is None:
					distribution_args = []

				distribution_kwargs = spec.distribution_kwargs
				if distribution_kwargs is None:
					distribution_kwargs = {}

				prior = distribution_class(
					parameter_name, *distribution_args, **distribution_kwargs
				)
				priors.append(prior)
			self.parameters = tuple(priors)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		distance = self.specification.distance
		if distance is None:
			distance = lambda e, _, simulated: simulated.item()  # noqa: E731

		def simulator_func(
			_: np.random.Generator, *parameter_values: list
		) -> np.ndarray:
			parameters = {}
			for i, name in enumerate(self.names):
				parameters[name] = parameter_values[i].item()

			abc_kwargs = self.get_calibration_func_kwargs()
			simulation_id = self.get_simulation_uuid()
			observed_data = self.specification.observed_data
			results = self.call_calibration_func(
				parameters, simulation_id, observed_data, **abc_kwargs
			)

			if not hasattr(results, "__iter__"):
				results = [results]
			if not isinstance(results, np.ndarray):
				results = np.array(results)

			return results

		with self.model:
			pm.Simulator(
				self.specification.experiment_name,
				simulator_func,
				params=self.parameters,
				distance=distance,
				sum_stat=self.specification.sum_stat,
				epsilon=self.specification.epsilon,
				observed=self.specification.observed_data,
			)

			method_kwargs = self.specification.method_kwargs
			if method_kwargs is None:
				method_kwargs = {}

			self.trace = pm.sample_smc(
				model=self.model,
				draws=self.specification.n_samples,
				chains=self.specification.n_chains,
				cores=self.specification.n_jobs,
				random_seed=self.specification.random_seed,
				**method_kwargs,
			)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		textsize = 7
		for plot in ["trace", "rank_vlines", "rank_bars"]:
			fig = az.plot_trace(
				self.trace, kind=plot, plot_kwargs={"textsize": textsize}
			)

			if outdir is not None:
				outfile = self.join(
					outdir, f"{time_now}-{task}-{experiment_name}-{plot}.png"
				)
				self.append_artifact(outfile)
				plt.tight_layout()
				plt.savefig(outfile)
				plt.close()
			else:
				fig.show()

		def _create_plot(plot_func: Callable, plot_kwargs: dict) -> None:
			plot_func(self.trace, **plot_kwargs)
			if outdir is not None:
				plot_name = plot_func.__name__.replace("_", "-")
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-{plot_name}.png",
				)
				self.append_artifact(outfile)
				plt.tight_layout()
				plt.savefig(outfile)
				plt.close()
			else:
				fig.show()

		_create_plot(
			az.plot_pair,
			plot_kwargs={
				"figsize": self.specification.figsize,
				"scatter_kwargs": dict(alpha=0.01),
				"marginals": True,
				"textsize": textsize,
			},
		)

		_create_plot(
			az.plot_violin,
			plot_kwargs={"figsize": self.specification.figsize, "textsize": textsize},
		)

		_create_plot(
			az.plot_posterior,
			plot_kwargs={"figsize": self.specification.figsize, "textsize": 5},
		)

		trace_summary_df = az.summary(self.trace).reset_index()
		trace_summary_df = trace_summary_df.rename(columns={"index": "name"})

		for row in trace_summary_df.to_dict("records"):
			name = row["name"]
			estimate = row["mean"]
			uncertainty = row["sd"]
			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(trace_summary_df, "trace-summary")

		trace_df = self.trace.to_dataframe(include_coords=False, groups="posterior")
		self.to_csv(trace_df, "trace")



"""Contains the implementations for the history matching methods

Implements the supported history matching methods.

"""

from collections.abc import Callable

import numpy as np
from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .ies_wrapper import IESHistoryMatching
from .pyesmda_wrapper import PyESMDAHistoryMatching

TASK = "history_matching"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	ies=IESHistoryMatching, pyesmda=PyESMDAHistoryMatching
)


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for history matching.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary of
			calibration implementations for history matching.
	"""
	return IMPLEMENTATIONS


class HistoryMatchingMethodModel(CalibrationModel):
	"""The history matching method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	covariance: np.ndarray | None = Field(
		description="The covariance matrix for variables", default=None
	)


class HistoryMatchingMethod(CalibrationMethodBase):
	"""The history matching method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: HistoryMatchingMethodModel,
		engine: str = "ies",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""HistoryMatchingMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (HistoryMatchingMethodModel): The calibration
				specification.
		    engine (str, optional): The history matching backend.
				Defaults to "ies".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains the surrogate modelling base class

The defined base class for surrogate modelling.

"""

import chaospy
import numpy as np

from .calibration_base import CalibrationWorkflowBase


class SurrogateBase(CalibrationWorkflowBase):
	"""The surrogate modelling base class."""

	def sample_parameters(self, n_samples: int) -> np.ndarray:
		"""Sample from the parameter space.

		Args:
			n_samples (int): The number of samples.

		Returns:
			np.ndarray: The sampled parameter values.
		"""
		X = self.parameters.sample(n_samples, rule="sobol").T
		n_replicates = self.specification.n_replicates
		if n_replicates > 1:
			X = np.repeat(X, n_replicates, axis=0)
			self.rng.shuffle(X)
		return X

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []
		parameters = []

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			data_type = spec.data_type
			self.data_types.append(data_type)

			distribution_name = (
				spec.distribution_name.replace("_", " ").title().replace(" ", "")
			)
			distribution_args = spec.distribution_args
			if distribution_args is None:
				distribution_args = []

			distribution_kwargs = spec.distribution_kwargs
			if distribution_kwargs is None:
				distribution_kwargs = {}

			dist_instance = getattr(chaospy, distribution_name)
			parameter = dist_instance(*distribution_args, **distribution_kwargs)
			parameters.append(parameter)

		self.parameters = chaospy.J(*parameters)



"""Contains the Scikit-Learn wrapper for GPyTorch

The defined Scikit-Learn wrapper for the GPyTorch library.

"""

import gpytorch
import torch
from skorch.probabilistic import ExactGPRegressor


class SingleTaskGPRegressionModel(gpytorch.models.ExactGP):
	"""The single task exact Gaussian process.

	Args:
	    gpytorch (ExactGP): The GPyTorch module.
	"""

	def __init__(
		self,
		likelihood: gpytorch.likelihoods.Likelihood,
		noise_init: float | None = None,
	) -> None:
		"""SingleTaskGPRegressionModel constructor.

		Args:
		    likelihood (gpytorch.likelihoods.Likelihood): The likelihood
		        for training the Gaussian process.
		    noise_init (float | None, optional): The
		        initial noise level. Defaults to None.
		"""
		super().__init__(train_inputs=None, train_targets=None, likelihood=likelihood)
		self.mean_module = gpytorch.means.ConstantMean()
		self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

	def forward(self, X: torch.Tensor) -> gpytorch.distributions.MultivariateNormal:
		"""The forward pass.

		Args:
		    X (torch.Tensor): The input tensor.

		Returns:
		    gpytorch.distributions.MultivariateNormal: The predicted
		        mean and covariance.
		"""
		mean_x = self.mean_module(X)
		covar_x = self.covar_module(X)
		return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


def get_single_task_exact_gp(
	lr: float, max_epochs: int, device: str = "cpu"
) -> ExactGPRegressor:
	"""Get an instance of a single task Gaussian process.

	Args:
	    lr (float): The learning rate.
	    max_epochs (int): The maximum number of epochs.
	    device (str, optional): The device to train the model. Defaults to "cpu".

	Returns:
	    ExactGPRegressor: The Gaussian process.
	"""
	return ExactGPRegressor(
		SingleTaskGPRegressionModel,
		optimizer=torch.optim.Adam,
		criterion=gpytorch.mlls.ExactMarginalLogLikelihood,
		lr=lr,
		max_epochs=max_epochs,
		device=device,
		batch_size=-1,
	)



"""Contains the implementations for the state estimation methods

Implements the supported state estimation methods.

"""

from collections.abc import Callable

from pydantic import Field

from ...base import CalibrationMethodBase, CalibrationWorkflowBase
from ...data_model import CalibrationModel
from .ensemble_kalman_filter import EKFStateEstimation

TASK = "state_estimation"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(ekf=EKFStateEstimation)


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for state estimation.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary of
			calibration implementations for state estimation.
	"""
	return IMPLEMENTATIONS


class StateEstimationMethodModel(CalibrationModel):
	"""The state estimation method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	replace_state_variables: bool = Field(
		description="Whether to replace or append the updated state variable",
		default=True,
	)
	stds: dict[str, float | list] | None = Field(
		description="The observation standard deviations", default=None
	)


class StateEstimationMethod(CalibrationMethodBase):
	"""The state estimation method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: StateEstimationMethodModel,
		engine: str = "ekf",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""StateEstimationMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (StateEstimationMethodModel): The calibration
				specification.
		    engine (str, optional): The state estimation backend.
				Defaults to "ekf".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains the implementations for Bayesian calibration methods using
OpenTurns

Implements the supported Bayesian calibration methods using
the OpenTurns library.

"""

import numpy as np
import openturns as ot
import openturns.viewer as viewer
import pandas as pd

from ..base import OpenTurnsBase
from ..data_model import ParameterEstimateModel


class OpenTurnsBayesianCalibration(OpenTurnsBase):
	"""The OpenTurns Bayesian calibration method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []
		self.bounds: tuple[list[float], list[float]] = ([], [])
		parameters = []

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			bounds = spec.distribution_bounds
			lower_bound, upper_bound = bounds
			lower_bounds, upper_bounds = self.bounds
			lower_bounds.append(lower_bound)
			upper_bounds.append(upper_bound)

			data_type = spec.data_type
			self.data_types.append(data_type)

			distribution_name = (
				spec.distribution_name.replace("_", " ").title().replace(" ", "")
			)

			distribution_args = spec.distribution_args
			if distribution_args is None:
				distribution_args = []

			distribution_kwargs = spec.distribution_kwargs
			if distribution_kwargs is None:
				distribution_kwargs = {}

			dist_instance = getattr(ot, distribution_name)
			parameter = dist_instance(*distribution_args, **distribution_kwargs)
			parameters.append(parameter)

		distribution_collection = ot.DistributionCollection(parameters)
		self.parameters = ot.JointDistribution(distribution_collection)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		bayesian_calibration_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: np.ndarray) -> np.ndarray:
			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				bayesian_calibration_kwargs,
			)
			if len(Y.shape) == 1:
				Y = np.expand_dims(Y, axis=1)
			return Y

		ot_func_wrapper = self.get_ot_func_wrapper(target_function)
		memoized_func = ot.MemoizeFunction(ot_func_wrapper)
		lower_bounds, upper_bounds = self.bounds
		support = ot.Interval(lower_bounds, upper_bounds)

		initial_state = self.specification.initial_state
		if initial_state is None:
			initial_state = []
			for i, lower_bound in enumerate(lower_bounds):
				upper_bound = upper_bounds[i]
				midpoint = np.median((lower_bound, upper_bound))
				initial_state.append(midpoint)

		self.sampler = ot.IndependentMetropolisHastings(
			memoized_func, support, initial_state, self.parameters
		)

		n_samples = self.specification.n_samples
		self.sample = self.sampler.getSample(n_samples)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		sample = np.array(self.sample)
		n_dim = self.parameters.getDimension()

		kernel = ot.KernelSmoothing()
		posterior = kernel.build(sample)
		grid = ot.GridLayout(n_dim, 1)
		grid.setTitle("Bayesian inference")

		for parameter_index in range(n_dim):
			graph = posterior.getMarginal(parameter_index).drawPDF()
			prior_graph = self.parameters.getMarginal(parameter_index).drawPDF()
			graph.add(prior_graph)
			parameter_name = self.names[parameter_index]
			graph.setTitle(parameter_name)
			graph.setLegends(["Posterior", "Prior"])
			grid.setGraph(parameter_index, 0, graph)
		view = viewer.View(grid, figure_kw={"figsize": self.specification.figsize})
		if outdir is not None:
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-plot-posterior.png"
			)
			self.append_artifact(outfile)
			view.save(outfile)

		trace_df = pd.DataFrame(sample, columns=self.names)
		for name in trace_df:
			estimate = trace_df[name].mean()
			uncertainty = trace_df[name].std()

			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(trace_df, "trace")



"""Contains the Scikit-Learn wrapper for Emukit

The defined Scikit-Learn wrapper for the Emukit library.

"""

import numpy as np
from emukit.model_wrappers import GPyModelWrapper
from GPy.models import GPRegression
from sklearn.base import BaseEstimator, MultiOutputMixin, RegressorMixin
from sklearn.metrics import r2_score
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y


class EmukitEstimator(MultiOutputMixin, RegressorMixin, BaseEstimator):
	def __init__(self, method_kwargs: dict | None = None):
		"""EmukitEstimator constructor.

		Args:
		    method_kwargs (dict, optional): The named arguments for the estimator.
		        Defaults to None.
		"""
		if method_kwargs is None:
			method_kwargs = {}
		self.method_kwargs = method_kwargs
		self.emulator = None

	def fit(self, X: np.ndarray, y: np.ndarray | None = None) -> "EmukitEstimator":
		"""Fit the estimator.

		Args:
		    X (np.ndarray): The simulation inputs.
		    y (np.ndarray | None, optional): The simulation outputs.
		        Defaults to None.

		Returns:
		    EmukitEstimator: The estimator.
		"""
		if len(y.shape) == 1:
			y = np.expand_dims(y, axis=1)

		X, y = check_X_y(
			X,
			y,
			multi_output=True,
			y_numeric=True,
			ensure_2d=True,
			dtype="numeric",
		)
		self.is_fitted_ = True
		self.n_features_in_ = X.shape[1]

		gp = GPRegression(X, y, **self.method_kwargs)
		self.emulator = GPyModelWrapper(gp)

		return self

	def predict(self, X: np.ndarray, return_std: bool = False) -> np.ndarray | tuple:
		"""Make a prediction.

		Args:
		    X (np.ndarray): The simulation inputs.
		    return_std (bool, optional): Whether to return the standard
				deviation. Defaults to False.

		Returns:
		    np.ndarray | tuple: The model predictions.
		"""
		check_is_fitted(self, "is_fitted_")
		X = check_array(X, ensure_2d=True, dtype="numeric")

		y_mu, y_var = self.emulator.predict(X)

		if return_std:
			return y_mu, np.sqrt(y_var)
		else:
			return y_mu

	def score(
		self, X: np.ndarray, y: np.ndarray, sample_weight: np.ndarray | None = None
	) -> float:
		"""Assess the estimator.

		Args:
		    X (np.ndarray): The simulation inputs.
		    y (np.ndarray): The simulation outputs.
		    sample_weight (np.ndarray | None, optional): Weighting factor
		        for the samples. Defaults to None.

		Returns:
		    float: The assessment score.
		"""
		y_pred = self.emulator.predict(X)
		return r2_score(y_pred, y, sample_weight=sample_weight)



"""Contains the implementations for the optimisation methods

Implements the supported optimisation methods.

"""

import importlib
from collections.abc import Callable

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .emukit_wrapper import EmukitOptimisation
from .openturns_wrapper import OpenTurnsOptimisation
from .optuna_wrapper import OptunaOptimisation

TASK = "optimisation"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	optuna=OptunaOptimisation,
	emukit=EmukitOptimisation,
	openturns=OpenTurnsOptimisation,
)

if importlib.util.find_spec("botorch") is not None:
	from .botorch_wrapper import BoTorchOptimisation

	IMPLEMENTATIONS["botorch"] = BoTorchOptimisation


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for optimisation.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]:
			The dictionary of calibration implementations for optimisation.
	"""
	return IMPLEMENTATIONS


class OptimisationMethodModel(CalibrationModel):
	"""The optimisation method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	directions: list[str] | None = Field(
		description="The list of objective directions", default=["minimize"]
	)
	acquisition_func: str | None = Field(
		description="The acquisition function for Bayesian optimisation", default="ei"
	)
	use_saasbo: bool = Field(
		description="Enable Sparse Axis-Aligned Subspace Bayesian Optimization",
		default=False,
	)


class OptimisationMethod(CalibrationMethodBase):
	"""The optimisation method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: OptimisationMethodModel,
		engine: str = "optuna",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""OptimisationMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (OptimisationMethodModel): The calibration
				specification.
		    engine (str, optional): The optimisation backend.
				Defaults to "optuna".
			implementation (CalibrationWorkflowBase | None): The calibration
				workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains the implementations for history matching methods using
pyESMDA

Implements the supported history matching methods using
the pyESMDA library.

"""

from collections.abc import Callable

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pyesmda import ESMDA, ESMDA_RS

from ..base import HistoryMatchingBase
from ..data_model import ParameterEstimateModel


def forward_model(
	m_ensemble: np.ndarray,
	workflow: HistoryMatchingBase,
	parameter_spec: dict,
	calibration_func: Callable,
	history_matching_kwargs: dict,
	observed_data: pd.DataFrame | np.ndarray,
	batched: bool = False,
) -> np.ndarray:
	"""The forward model for the ensemble simulation.

	Args:
		m_ensemble (np.ndarray): The ensemble simulation parameters.
		workflow (HistoryMatchingBase) The calibration workflow.
		parameter_spec (dict): The parameter specification.
		calibration_func (Callable): The history matching function.
		history_matching_kwargs (dict): Named arguments for the
			history matching function.
		observed_data (pd.DataFrame | np.ndarray): The observed data.
		batched (bool, optional): Whether to batch the history
			matching function. Defaults to False.

	Returns:
		np.ndarray: The ensemble results.
	"""
	parameters = []
	for i in range(m_ensemble.shape[0]):
		parameter_set = {}
		for k in parameter_spec:
			parameter_set[k] = parameter_spec[k][i]
		parameters.append(parameter_set)

	if history_matching_kwargs is None:
		history_matching_kwargs = {}

	simulation_ids = [
		workflow.get_simulation_uuid() for _ in range(m_ensemble.shape[0])
	]
	if batched:
		ensemble_outputs = calibration_func(
			parameters, simulation_ids, observed_data, **history_matching_kwargs
		)
	else:
		ensemble_outputs = []
		for i, parameter in enumerate(parameters):
			simulation_id = simulation_ids[i]
			outputs = calibration_func(
				parameter, simulation_id, observed_data, **history_matching_kwargs
			)
			ensemble_outputs.append(outputs)

	ensemble_outputs = np.array(ensemble_outputs)
	return ensemble_outputs


class PyESMDAHistoryMatching(HistoryMatchingBase):
	"""The pyESMDA history matching method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		smoother_name = self.specification.method
		smoothers = dict(esmda=ESMDA, esmda_rs=ESMDA_RS)
		smoother_class = smoothers.get(smoother_name, None)
		if smoother_class is None:
			raise ValueError(
				f"Unsupported ensemble smoother: {smoother_name}.",
				f"Supported ensemble smoothers are {', '.join(smoothers)}",
			)

		n_jobs = self.specification.n_jobs
		if n_jobs > 1:
			is_parallel_analyse_step = True
		else:
			is_parallel_analyse_step = False

		observed_data = self.specification.observed_data
		cov_obs = self.specification.covariance
		if cov_obs is None:
			cov_obs = np.eye(observed_data.shape[0])

		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		m_init = self.specification.X
		if m_init is None:
			m_init = [self.parameters[k] for k in self.parameters]
			m_init = np.array(m_init).T

		if smoother_name == "esmda":
			method_kwargs["n_assimilations"] = self.specification.n_iterations

		history_matching_kwargs = self.get_calibration_func_kwargs()
		self.solver = smoother_class(
			obs=observed_data,
			m_init=m_init,
			forward_model=forward_model,
			forward_model_kwargs=dict(
				workflow=self,
				parameter_spec=self.parameters,
				calibration_func=self.call_calibration_func,
				history_matching_kwargs=history_matching_kwargs,
				observed_data=observed_data,
				batched=self.specification.batched,
			),
			cov_obs=cov_obs,
			random_state=self.specification.random_seed,
			batch_size=n_jobs,
			is_parallel_analyse_step=is_parallel_analyse_step,
			**method_kwargs,
		)

		self.solver.solve()

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		smoother_name = self.specification.method
		n_iterations = self.specification.n_iterations
		n_samples = self.specification.n_samples

		means = np.average(self.solver.m_prior, axis=0)
		stds = np.sqrt(np.diagonal(self.solver.cov_mm))

		parameter_names = list(self.parameters.keys())
		parameter_samples = {}

		fig, axes = plt.subplots(
			nrows=len(parameter_names), figsize=self.specification.figsize
		)
		if not isinstance(axes, np.ndarray):
			axes = [axes]

		for i, parameter_name in enumerate(parameter_names):
			axes[i].set_title(parameter_name)
			axes[i].hist(self.parameters[parameter_name], label="Prior")
			mu = means[i]
			sigma = stds[i]
			parameter_samples[parameter_name] = self.rng.normal(
				mu, sigma, size=n_samples
			)
			axes[i].hist(
				parameter_samples[parameter_name],
				label=f"{smoother_name} ({n_iterations}) Posterior",
				alpha=0.5,
			)
			axes[i].legend()
		self.present_fig(fig, outdir, time_now, task, experiment_name, "plot-slice")

		pred_dfs = [pd.DataFrame(preds) for preds in self.solver.d_pred]
		output_labels = self.specification.output_labels
		if output_labels is None:
			output_labels = ["output"]
		output_label = output_labels[0]

		observed_data = self.specification.observed_data
		fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)
		X = np.arange(0, observed_data.shape[0], 1)
		axes[0].plot(X, observed_data)
		axes[0].set_title(f"Observed {output_label}")

		for pred_df in pred_dfs:
			pred_df[0].plot(ax=axes[1])
		axes[1].set_title(f"Ensemble {output_label}")
		fig.tight_layout()
		self.present_fig(
			fig, outdir, time_now, task, experiment_name, f"ensemble-{output_label}"
		)

		X_IES_df = pd.DataFrame(parameter_samples)
		for name in X_IES_df:
			estimate = X_IES_df[name].mean()
			uncertainty = X_IES_df[name].std()

			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(X_IES_df, "posterior")

		for pred_df in pred_dfs:
			pred_df["x"] = pred_df.index
		pred_df = pd.concat(pred_dfs)
		pred_df.columns = [output_label, "x"]
		self.to_csv(pred_df, f"ensemble-{output_label}")



"""Contains the implementations for the evolutionary algorithm methods

Implements the supported evolutionary algorithm methods.

"""

import importlib
from collections.abc import Callable
from typing import Any

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .spotpy_wrapper import SPOTPYEvolutionary

TASK = "evolutionary"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	spotpy=SPOTPYEvolutionary
)

if importlib.util.find_spec("evotorch") is not None:
	from .evotorch_wrapper import EvoTorchEvolutionary

	IMPLEMENTATIONS["evotorch"] = EvoTorchEvolutionary


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for evolutionary algorithm.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary
			of calibration implementations for evolutionary algorithm.
	"""
	return IMPLEMENTATIONS


class EvolutionaryMethodModel(CalibrationModel):
	"""The evolutionary algorithm method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	objective: str | None = Field(description="The objective function", default="rmse")
	directions: list[str] | None = Field(
		description="The list of objective directions", default=["minimize"]
	)
	operators: dict[str, Any] | None = Field(
		description="The dictionary of evolutionary operators", default=None
	)


class EvolutionaryMethod(CalibrationMethodBase):
	"""The evolutionary algorithm method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: EvolutionaryMethodModel,
		engine: str = "spotpy",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""EvolutionaryMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (EvolutionaryMethodModel): The calibration
				specification.
		    engine (str, optional): The evolutionary algorithm backend.
				Defaults to "spotpy".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains the implementations for evolutionary algorithms using SPOTPY

Implements the supported evolutionary algorithms using the SPOTPY library.

"""

from collections.abc import Callable

import numpy as np
import pandas as pd
import spotpy.analyser as analyser
import spotpy.likelihoods as likelihoods
import spotpy.objectivefunctions as objectivefunctions
import spotpy.parameter
from spotpy.algorithms import NSGAII, abc, demcz, dream, fscabc, sceua
from spotpy.parameter import Base, generate

from ..base import CalibrationWorkflowBase
from ..data_model import ParameterEstimateModel


class SPOTSetup:
	"""The SPOTPY calibration setup."""

	def __init__(
		self,
		workflow: CalibrationWorkflowBase,
		objective_function: Callable,
		evolutionary_name: str,
	):
		"""SPOTSetup constructor.

		Args:
			workflow (CalibrationWorkflowBase): The calibration workflow object.
			objective_function (Callable): The simulation objective function.
			evolutionary_name (str): The name of the evolutionary algorithm.
		"""
		self.objective_function = objective_function
		self.evolutionary_name = evolutionary_name
		self.setup_from_workflow(workflow)

	def setup_from_workflow(self, workflow: CalibrationWorkflowBase) -> None:
		"""Configure the calibration procedure from the workflow object.

		Args:
			workflow (CalibrationWorkflowBase): The calibration workflow object.
		"""
		parameter_names = []
		data_types = []
		priors: list[Base] = []

		parameter_spec = workflow.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			parameter_names.append(parameter_name)
			data_type = spec.data_type
			data_types.append(data_type)

			distribution_name = (
				spec.distribution_name.replace("_", " ").title().replace(" ", "")
			)
			distribution_class = getattr(spotpy.parameter, distribution_name)
			distribution_args = spec.distribution_args
			if distribution_args is None:
				distribution_args = []

			distribution_kwargs = spec.distribution_kwargs
			if distribution_kwargs is None:
				distribution_kwargs = {}

			prior = distribution_class(
				parameter_name, *distribution_args, **distribution_kwargs
			)
			priors.append(prior)

		self.parameter_names = parameter_names
		self.data_types = data_types
		self.priors = priors
		self.call_calibration_func = workflow.call_calibration_func
		self.observed_data = workflow.specification.observed_data
		self.evolutionary_kwargs = workflow.get_calibration_func_kwargs()
		self.workflow = workflow

	def parameters(self) -> np.ndarray:
		"""Generate parameters from the prior specification.

		Returns:
			np.ndarray: The generated parameters.
		"""
		return generate(self.priors)

	def simulation(self, X: np.ndarray) -> np.ndarray:
		"""Run the simulation.

		Args:
			X (np.ndarray): The simulation parameter vector.

		Returns:
			np.ndarray: The simulation results.
		"""
		X = [X]
		results = self.workflow.calibration_func_wrapper(
			X,
			self.workflow,
			self.observed_data,
			self.parameter_names,
			self.data_types,
			self.evolutionary_kwargs,
		)
		return results[0]

	def evaluation(self) -> np.ndarray | pd.DataFrame:
		"""Get the observed data.

		Returns:
			np.ndarray | pd.DataFrame: The observed data.
		"""
		return self.observed_data

	def objectivefunction(
		self,
		simulation: np.ndarray | pd.DataFrame,
		evaluation: np.ndarray | pd.DataFrame,
	) -> float:
		"""Call the objective function on simulated and observed data.

		Args:
			simulation (np.ndarray | pd.DataFrame): The simulated data.
			evaluation (np.ndarray | pd.DataFrame): The observed data.

		Returns:
			float: The objective function results.
		"""
		if self.evolutionary_name == "dream":
			objective = self.objective_function(evaluation, simulation)
		else:
			objective = -self.objective_function(evaluation, simulation)
		return objective


class SPOTPYEvolutionary(CalibrationWorkflowBase):
	"""The SPOTPY evolutionary algorithm method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		evolutionary_name = self.specification.method
		objective = self.specification.objective

		if evolutionary_name == "dream":
			objective_function = getattr(likelihoods, objective)
		else:
			objective_function = getattr(objectivefunctions, objective)

		self.spot_setup = SPOTSetup(self, objective_function, evolutionary_name)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		evolutionary_name = self.specification.method
		evolutionary_algorithms = dict(
			abc=abc, demcz=demcz, dream=dream, fscabc=fscabc, nsgaii=NSGAII, sceua=sceua
		)
		evolutionary_class = evolutionary_algorithms.get(evolutionary_name, None)
		if evolutionary_class is None:
			raise ValueError(
				f"Unsupported evolutionary algorithm: {evolutionary_name}.",
				f"Supported algorithms are {', '.join(evolutionary_algorithms.keys())}",
			)

		_, time_now, dbname, outdir = self.prepare_analyze()
		if outdir is None:
			dbformat = "ram"
		else:
			dbformat = "csv"
			dbname = self.join(outdir, f"{time_now}_{dbname}")

		self.sampler = evolutionary_class(
			spot_setup=self.spot_setup,
			dbname=dbname,
			dbformat=dbformat,
			parallel="seq",
			save_sim=True,
		)

		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}
		n_samples = self.specification.n_samples
		self.sample_results = self.sampler.sample(n_samples, **method_kwargs)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		results = self.sampler.getdata()

		parameter_values = {}
		for col in results.dtype.names:
			if not col.startswith("par"):
				continue
			values = np.array(results[col])
			parameter_values[col] = values

			name = col.replace("par", "")
			estimate = values.mean()
			uncertainty = values.std()

			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		trace_df = pd.DataFrame(parameter_values)
		self.to_csv(trace_df, "trace")

		for plot_func in [
			analyser.plot_fast_sensitivity,
			analyser.plot_parametertrace,
			analyser.plot_parameterInteraction,
		]:
			plot_func_name = plot_func.__name__.replace("_", "-")
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-{plot_func_name}.png"
			)
			self.append_artifact(outfile)
			plot_func(results, fig_name=outfile)

		evaluation = self.specification.observed_data
		for plot_func in [
			analyser.plot_objectivefunction,
			analyser.plot_regression,
		]:
			plot_func_name = plot_func.__name__.replace("_", "-")
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-{plot_func_name}.png"
			)
			self.append_artifact(outfile)
			plot_func(results, evaluation, fig_name=outfile)

		if self.specification.method == "dream":
			plot_func = analyser.plot_gelman_rubin
			plot_func_name = plot_func.__name__.replace("_", "-")
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-{plot_func_name}.png"
			)
			self.append_artifact(outfile)
			plot_func(results, self.sample_results, outfile)



"""Contains the implementations for the Bayesian calibration methods

Implements the supported Bayesian calibration methods.

"""

from collections.abc import Callable

from pydantic import Field

from ..base import CalibrationMethodBase, CalibrationWorkflowBase
from ..data_model import CalibrationModel
from .dynesty_wrapper import DynestyBayesianCalibration
from .emcee_wrapper import EmceeBayesianCalibration
from .openturns_wrapper import OpenTurnsBayesianCalibration

TASK = "bayesian_calibration"
IMPLEMENTATIONS: dict[str, type[CalibrationWorkflowBase]] = dict(
	openturns=OpenTurnsBayesianCalibration,
	emcee=EmceeBayesianCalibration,
	dynesty=DynestyBayesianCalibration,
)


def get_implementations() -> dict[str, type[CalibrationWorkflowBase]]:
	"""Get the calibration implementations for Bayesian calibration.

	Returns:
		Dict[str, type[CalibrationWorkflowBase]]: The dictionary of
			calibration implementations for Bayesian calibration.
	"""
	return IMPLEMENTATIONS


class BayesianCalibrationMethodModel(CalibrationModel):
	"""The Bayesian calibration method data model.

	Args:
	    BaseModel (CalibrationModel): The calibration base model class.
	"""

	log_density: bool = Field(
		description="Take the log of the target density.", default=False
	)
	initial_state: list | bool = Field(
		description="Initial state of the chain.", default=None
	)
	moves: dict[str, float] = Field(
		description="List of methods for updating coordinates of ensemble walkers.",
		default=None,
	)


class BayesianCalibrationMethod(CalibrationMethodBase):
	"""The Bayesian calibration method class."""

	def __init__(
		self,
		calibration_func: Callable,
		specification: BayesianCalibrationMethodModel,
		engine: str = "openturns",
		implementation: CalibrationWorkflowBase | None = None,
	) -> None:
		"""BayesianCalibrationMethod constructor.

		Args:
			calibration_func (Callable): The calibration function.
				For example, a simulation function or objective function.
		    specification (BayesianCalibrationMethodModel): The
				calibration specification.
		    engine (str, optional): The Bayesian calibration
				backend. Defaults to "openturns".
			implementation (CalibrationWorkflowBase | None): The
				calibration workflow implementation.
		"""
		super().__init__(
			calibration_func,
			specification,
			TASK,
			engine,
			IMPLEMENTATIONS,
			implementation,
		)



"""Contains the implementations for history matching methods using
iterative_ensemble_smoother

Implements the supported history matching methods using
the iterative_ensemble_smoother library.

"""

import iterative_ensemble_smoother as ies
import numpy as np
import pandas as pd
from iterative_ensemble_smoother.utils import steplength_exponential
from matplotlib import pyplot as plt

from ..base import HistoryMatchingBase
from ..data_model import ParameterEstimateModel


class IESHistoryMatching(HistoryMatchingBase):
	"""The iterative_ensemble_smoother history matching method class."""

	def convert_parameters(self, X: np.ndarray) -> list[dict[str, float]]:
		"""Convert the parameters from an array to a list of records.

		Args:
		    X (np.ndarray): The array of parameters.

		Returns:
		    List[Dict[str, float]]: The list of parameters.
		"""
		parameters = []
		ensemble_size = self.specification.n_samples
		parameter_spec = self.specification.parameter_spec.parameters
		for i in range(ensemble_size):
			parameter_set = {}
			for j, spec in enumerate(parameter_spec):  # type: ignore[arg-type]
				parameter_name = spec.name
				parameter_set[parameter_name] = X[j][i]
			parameters.append(parameter_set)
		return parameters

	def run_simulation(self, parameters: list[dict[str, float]]) -> np.ndarray:
		"""Run the simulation for the history matching procedure.

		Args:
		    parameters (List[Dict[str, float]]): The list of
				simulation parameters.

		Returns:
		    np.ndarray: The ensemble outputs.
		"""
		observed_data = self.specification.observed_data
		history_matching_kwargs = self.get_calibration_func_kwargs()

		simulation_ids = [self.get_simulation_uuid() for _ in range(len(parameters))]

		if self.specification.batched:
			ensemble_outputs = self.call_calibration_func(
				parameters, simulation_ids, observed_data, **history_matching_kwargs
			)
		else:
			ensemble_outputs = []
			for i, parameter in enumerate(parameters):
				simulation_id = simulation_ids[i]
				outputs = self.call_calibration_func(
					parameter, simulation_id, observed_data, **history_matching_kwargs
				)
				ensemble_outputs.append(outputs)  # type: ignore[arg-type]

		ensemble_outputs = np.array(ensemble_outputs).T
		return ensemble_outputs

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		parameters = self.specification.X
		if parameters is None:
			parameters = []
			ensemble_size = self.specification.n_samples
			for i in range(ensemble_size):
				parameter_set = {}
				for k in self.parameters:
					parameter_set[k] = self.parameters[k][i]
				parameters.append(parameter_set)

		ensemble_outputs = self.specification.Y
		if ensemble_outputs is None:
			ensemble_outputs = self.run_simulation(parameters)

		smoother_name = self.specification.method
		smoothers = dict(sies=ies.SIES, esmda=ies.ESMDA)
		smoother_class = smoothers.get(smoother_name, None)
		if smoother_class is None:
			raise ValueError(
				f"Unsupported iterative ensemble smoother: {smoother_name}.",
				f"Supported iterative ensemble smoothers are {', '.join(smoothers)}",
			)

		param_values = np.array([distr for distr in self.parameters.values()])
		X_i = param_values.copy()

		n_iterations = self.specification.n_iterations
		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}
		if smoother_name == "sies":
			method_kwargs["parameters"] = X_i
		else:
			method_kwargs["alpha"] = n_iterations

		observations = self.specification.observed_data
		covariance = self.specification.covariance
		if covariance is None:
			covariance = np.eye(observations.shape[0])

		smoother = smoother_class(
			covariance=covariance,
			observations=observations,
			seed=self.specification.random_seed,
			**method_kwargs,
		)

		Y_i = ensemble_outputs.copy()
		if smoother_name == "sies":
			for i, alpha_i in enumerate(range(n_iterations)):
				if self.specification.verbose:
					print(f"SIES iteration {i + 1}/{n_iterations}")
				step_length = steplength_exponential(i + 1)
				X_i = smoother.sies_iteration(Y_i, step_length=step_length)
				parameters = self.convert_parameters(X_i)
				Y_i = self.run_simulation(parameters)
		else:
			for i, alpha_i in enumerate(smoother.alpha):
				if self.specification.verbose:
					print(
						f"ESMDA iteration {i + 1}/{smoother.num_assimilations()}"
						+ f" with inflation factor alpha_i={alpha_i}"
					)

				X_i = smoother.assimilate(X_i, Y=Y_i)
				parameters = self.convert_parameters(X_i)
				Y_i = self.run_simulation(parameters)

		self.X_IES = X_i
		self.Y_IES = Y_i

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		smoother_name = self.specification.method
		n_iterations = self.specification.n_iterations

		parameter_names = list(self.parameters.keys())
		fig, axes = plt.subplots(
			nrows=len(parameter_names), figsize=self.specification.figsize
		)
		if not isinstance(axes, np.ndarray):
			axes = [axes]

		for i, parameter_name in enumerate(parameter_names):
			axes[i].set_title(parameter_name)
			axes[i].hist(self.parameters[parameter_name], label="Prior")
			axes[i].hist(
				self.X_IES[i, :],
				label=f"{smoother_name} ({n_iterations}) Posterior",
				alpha=0.5,
			)
			axes[i].legend()
		self.present_fig(fig, outdir, time_now, task, experiment_name, "plot-slice")

		ensemble_size = self.specification.n_samples
		output_labels = self.specification.output_labels
		if output_labels is None:
			output_labels = ["output"]
		output_label = output_labels[0]

		X = np.arange(0, self.Y_IES.shape[0], 1)
		fig, axes = plt.subplots(nrows=2, figsize=self.specification.figsize)

		observed_data = self.specification.observed_data
		axes[0].plot(X, observed_data)
		axes[0].set_title(f"Observed {output_label}")

		for i in range(ensemble_size):
			axes[1].plot(X, self.Y_IES.T[i])
		axes[1].set_title(f"Ensemble {output_label}")
		self.present_fig(
			fig, outdir, time_now, task, experiment_name, f"ensemble-{output_label}"
		)

		X_IES_df = pd.DataFrame(self.X_IES.T, columns=parameter_names)
		for name in X_IES_df:
			estimate = X_IES_df[name].mean()
			uncertainty = X_IES_df[name].std()

			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(X_IES_df, "posterior")

		Y_IES_df = pd.DataFrame(
			self.Y_IES.T,
			columns=[f"{output_label}_{i + 1}" for i in range(self.Y_IES.shape[0])],
		)
		self.to_csv(Y_IES_df, f"ensemble-{output_label}")



"""Contains the implementations for reliability analysis methods using
OpenTurns

Implements the supported reliability analysis methods using
the OpenTurns library.

"""

import numpy as np
import openturns as ot
import openturns.viewer as viewer

from ..base import OpenTurnsBase


class OpenTurnsReliabilityAnalysis(OpenTurnsBase):
	"""The OpenTurns reliability analysis method class."""

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		method = self.specification.method
		supported_methods = ["monte_carlo", "sobol", "subset"]
		if method not in supported_methods:
			raise ValueError(
				f"Unsupported sampler: {method}.",
				f"Supported samplers: {', '.join(supported_methods)}",
			)

		comparison = self.specification.comparison
		valid_comparisons = [
			"less",
			"less_or_equal",
			"greater",
			"greater_or_equal",
			"equal",
		]
		if comparison not in valid_comparisons:
			raise ValueError(
				f"Unsupported comparison: {comparison}.",
				f"Supported comparisons: {', '.join(valid_comparisons)}",
			)

		reliability_kwargs = self.get_calibration_func_kwargs()

		def target_function(X: np.ndarray) -> np.ndarray:
			Y = self.calibration_func_wrapper(
				X,
				self,
				self.specification.observed_data,
				self.names,
				self.data_types,
				reliability_kwargs,
			)
			if len(Y.shape) == 1:
				Y = np.expand_dims(Y, axis=1)
			return Y

		self.input_vector = ot.RandomVector(self.parameters)
		ot_func_wrapper = self.get_ot_func_wrapper(target_function)
		memoized_func = ot.MemoizeFunction(ot_func_wrapper)
		G = ot.CompositeRandomVector(memoized_func, self.input_vector)

		threshold = self.specification.threshold
		comparison = comparison.replace("_", " ").title().replace(" ", "")
		comparison_func = getattr(ot, comparison)

		event = ot.ThresholdEvent(G, comparison_func(), threshold)

		if method == "monte_carlo":
			self.experiment = ot.MonteCarloExperiment()
		elif method == "sobol":
			sequence = ot.SobolSequence()
			self.experiment = ot.LowDiscrepancyExperiment(sequence, 1)
			self.experiment.setRandomize(True)

		if method == "subset":
			self.sampler = ot.SubsetSampling(event)
			self.sampler.setKeepSample(True)
		else:
			self.sampler = ot.ProbabilitySimulationAlgorithm(event, self.experiment)
			self.sampler.setMaximumCoefficientOfVariation(0.05)

		n_samples = self.specification.n_samples
		n_jobs = self.specification.n_jobs

		self.sampler.setMaximumOuterSampling(n_samples)
		self.sampler.setBlockSize(n_jobs)
		self.sampler.run()

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		method = self.specification.method

		result = self.sampler.getResult()
		graph = result.drawImportanceFactors()
		view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
		if outdir is not None:
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-{method}-plot-importance-factors.png",
			)
			self.append_artifact(outfile)
			view.save(outfile)

		graph = self.sampler.drawProbabilityConvergence()
		graph.setLogScale(ot.GraphImplementation.LOGX)
		view = viewer.View(graph, figure_kw={"figsize": self.specification.figsize})
		if outdir is not None:
			outfile = self.join(
				outdir,
				f"{time_now}-{task}-{experiment_name}-{method}-plot-probability-convergence.png",
			)
			self.append_artifact(outfile)
			view.save(outfile)



"""Contains the implementations for sensitivity analysis methods using SALib

Implements the supported sensitivity analysis methods using the SALib library.

"""

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from SALib import ProblemSpec

from ..base import CalibrationWorkflowBase


class SALibSensitivityAnalysis(CalibrationWorkflowBase):
	"""The SALib sensitivity analysis method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []
		bounds = []
		dists = []

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			data_type = spec.data_type
			self.data_types.append(data_type)

			dists.append("unif")

			lower_bound, upper_bound = self.get_parameter_bounds(spec)
			bounds.append([lower_bound, upper_bound])

		problem = {
			"num_vars": len(self.names),
			"names": self.names,
			"bounds": bounds,
			"dists": dists,
			"groups": self.specification.groups,
			"outputs": self.specification.output_labels,
		}

		self.sp = ProblemSpec(problem)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		data_types = []
		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			data_type = spec.data_type
			data_types.append(data_type)

		sampler_name = self.specification.method
		sample_func = getattr(self.sp, f"sample_{sampler_name}")
		sampler_kwargs = self.specification.method_kwargs
		if sampler_kwargs is None:
			sampler_kwargs = {}
		sampler_kwargs["seed"] = self.specification.random_seed
		n_samples = self.specification.n_samples

		sp_samples = self.specification.X
		sp_results = self.specification.Y
		if sp_samples is None:
			sample_func(n_samples, **sampler_kwargs)
			n_replicates = self.specification.n_replicates
			if n_replicates > 1 and sp_results is None:
				X = self.sp.samples
				X = np.repeat(X, n_replicates, axis=0)
				self.rng.shuffle(X)
				self.sp.samples = X
		else:
			self.sp.samples = sp_samples

		sensitivity_kwargs = self.get_calibration_func_kwargs()

		n_jobs = self.specification.n_jobs
		if sp_results is None:
			if n_jobs == 1:
				self.sp.evaluate(
					self.calibration_func_wrapper,
					self,
					self.specification.observed_data,
					self.names,
					self.data_types,
					sensitivity_kwargs,
				)
			else:
				self.sp.evaluate_parallel(
					self.calibration_func_wrapper,
					self,
					self.specification.observed_data,
					self.names,
					self.data_types,
					sensitivity_kwargs,
					nprocs=n_jobs,
				)
		else:
			self.sp.results = sp_results
		analyze_func = getattr(self.sp, f"analyze_{sampler_name}")
		analyze_kwargs = self.specification.analyze_kwargs
		if analyze_kwargs is None:
			analyze_kwargs = {}
		analyze_kwargs["seed"] = self.specification.random_seed
		analyze_func(**analyze_kwargs)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()
		sampler_name = self.specification.method

		self.sp.plot()
		plt.tight_layout()
		if outdir is not None:
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-indices.png"
			)
			self.append_artifact(outfile)
			plt.savefig(outfile)
		else:
			plt.show()
		plt.close()

		self.sp.heatmap()
		plt.tight_layout()
		if outdir is not None:
			outfile = self.join(
				outdir, f"{time_now}-{task}-{experiment_name}-heatmap.png"
			)
			self.append_artifact(outfile)
			plt.savefig(outfile)
		else:
			plt.show()
		plt.close()

		if outdir is None:
			return

		def recursive_write_csv(dfs: pd.DataFrame) -> None:
			if isinstance(dfs, list):
				for df in dfs:
					recursive_write_csv(df)
			else:
				si_df = dfs.reset_index().rename(columns={"index": "parameter"})
				si_type = si_df.columns[1]
				self.to_csv(si_df, si_type)

		si_dfs = self.sp.to_df()
		if isinstance(si_dfs, list):
			recursive_write_csv(si_dfs)
		else:
			si_df = si_dfs.reset_index().rename(columns={"index": "parameter"})
			self.to_csv(si_df, sampler_name)



"""Contains utilities for the design of experiments.

This module defines utility functions for various
experimental design methods.
"""

from itertools import product

import numpy as np

from ..data_model import ParameterSpecification


def get_full_factorial_design(
	parameter_spec: ParameterSpecification,
) -> np.ndarray:  # pragma: no cover
	"""Get a full factorial design from a parameter specification.

	Args:
		parameter_spec (ParameterSpecification): The parameter specification.

	Returns:
		np.ndarray: The full factorial design.
	"""
	factorial_design = []
	for parameter in parameter_spec.parameters:
		parameter_values = parameter.parameter_values
		if parameter_values is not None:
			factorial_design.append(parameter_values)

	cartesian_product = list(product(*factorial_design))
	cartesian_product = [np.array(combination) for combination in cartesian_product]
	return np.array(cartesian_product)



import importlib

from .calibration_base import CalibrationMethodBase, CalibrationWorkflowBase
from .emukit_base import EmukitBase
from .example_model_base import ExampleModelBase, ExampleModelContainer
from .history_matching_base import HistoryMatchingBase
from .openturns_base import OpenTurnsBase
from .surrogate_base import SurrogateBase

__all__ = [
	CalibrationMethodBase,
	CalibrationWorkflowBase,
	EmukitBase,
	ExampleModelBase,
	ExampleModelContainer,
	HistoryMatchingBase,
	OpenTurnsBase,
	SurrogateBase,
]

if importlib.util.find_spec("torch") is not None:
	from .sbi_base import SimulationBasedInferenceBase

	__all__.append(SimulationBasedInferenceBase)



"""Contains the ELFI base class

The defined base class for the ELFI library.

"""

from collections.abc import Callable

import elfi
import numpy as np

from ...base import CalibrationWorkflowBase


class ELFIBase(CalibrationWorkflowBase):
	"""The ELFI base class."""

	def dist_name_processing(self, name: str) -> str:
		"""Apply data preprocessing to the distribution name.

		Args:
			name (str): The unprocessed distribution name.

		Returns:
			str: The processed distribution name.
		"""
		name = name.replace(" ", "_").lower()

		if name == "normal":
			name = "norm"
		return name

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.bounds = {}
		self.priors = []
		parameter_spec = self.specification.parameter_spec.parameters

		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			distribution_name = self.dist_name_processing(spec.distribution_name)

			bounds = spec.distribution_bounds
			if bounds is not None:
				self.bounds[parameter_name] = bounds

			distribution_args = spec.distribution_args
			if distribution_args is None:
				distribution_args = []

			distribution_kwargs = spec.distribution_kwargs
			if distribution_kwargs is None:
				distribution_kwargs = {}

			prior = elfi.Prior(
				distribution_name,
				*distribution_args,
				**distribution_kwargs,
				name=parameter_name,
			)
			self.priors.append(prior)

	def create_simulator(
		self, simulator_func: Callable, take_log: bool = False
	) -> None:
		"""Create a simulator object.

		Args:
			simulator_func (Callable): The simulator function.
			take_log (bool, optional): Whether to take the log distance.
				Defaults to False.
		"""
		simulator = elfi.Simulator(
			simulator_func,
			*self.priors,
			observed=self.specification.observed_data,
			name="simulator",
		)
		identity_statistic = elfi.Summary(
			lambda y: y, simulator, name="identity_statistic"
		)

		distance = elfi.Distance(
			lambda simulated, _: simulated, identity_statistic, name="identity_distance"
		)
		if take_log:
			distance = elfi.Operation(np.log, distance, name="log_distance")

		self.distance = distance



from .implementation import (
	BayesianCalibrationMethod,
	BayesianCalibrationMethodModel,
	get_implementations,
)

_all__ = [
	BayesianCalibrationMethod,
	BayesianCalibrationMethodModel,
	get_implementations,
]



"""Contains the implementations for Bayesian calibration methods using
dynesty

Implements the supported Bayesian calibration methods using
the dynesty library.

"""

import sys

import dynesty
import numpy as np
import pandas as pd
from dynesty import plotting as dyplot
from dynesty import utils as dyfunc
from matplotlib import pyplot as plt

from ..base import CalibrationWorkflowBase
from ..data_model import ParameterDataType, ParameterEstimateModel


def prior_transform(utheta: list[float], self: CalibrationWorkflowBase) -> list[float]:
	transformed_theta = []
	for i, bounds in enumerate(self.bounds):
		x = utheta[i]
		lower_bound, upper_bound = bounds
		x = lower_bound + (upper_bound - lower_bound) * x
		data_type = self.data_types[i]
		if data_type == ParameterDataType.DISCRETE:
			x = int(x)
		transformed_theta.append(x)
	return transformed_theta


def target_function(
	X: np.ndarray, self: CalibrationWorkflowBase, bayesian_calibration_kwargs: dict
) -> np.ndarray:
	Y = self.calibration_func_wrapper(
		[X],
		self,
		self.specification.observed_data,
		self.names,
		self.data_types,
		bayesian_calibration_kwargs,
	)
	if np.isscalar(Y):
		return Y
	elif len(Y) == 1:
		return Y.item()
	else:
		raise ValueError(
			"Log likelihood for DynestyBayesianCalibration must be scalar value."
		)


class DynestyBayesianCalibration(CalibrationWorkflowBase):
	"""The dynesty Bayesian calibration method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		self.names = []
		self.data_types = []
		self.bounds = []

		parameter_spec = self.specification.parameter_spec.parameters
		for spec in parameter_spec:
			parameter_name = spec.name
			self.names.append(parameter_name)

			bounds = self.get_parameter_bounds(spec)
			self.bounds.append(bounds)

			data_type = spec.data_type
			self.data_types.append(data_type)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		sampler_name = self.specification.method
		supported_samplers = dict(
			dynamic=dynesty.DynamicNestedSampler,
			nested=dynesty.NestedSampler,
		)
		sampler_class = supported_samplers.get(sampler_name, None)
		if sampler_class is None:
			raise ValueError(
				f"Unsupported dynesty sampler: {sampler_name}.",
				f"Supported dynesty samplers are {', '.join(supported_samplers)}",
			)

		nlive = self.specification.n_init
		maxiter = self.specification.n_iterations
		if maxiter is None:
			maxiter = sys.maxsize
		maxcall = self.specification.n_samples
		if maxcall is None:
			maxcall = sys.maxsize

		random_seed = self.specification.random_seed
		rstate = np.random.default_rng(random_seed)
		n_jobs = self.specification.n_jobs

		bayesian_calibration_kwargs = self.get_calibration_func_kwargs()
		method_kwargs = self.specification.method_kwargs
		if method_kwargs is None:
			method_kwargs = {}

		if n_jobs > 1:
			with dynesty.pool.Pool(
				n_jobs,
				target_function,
				prior_transform,
				logl_args=(self, bayesian_calibration_kwargs),
				ptform_args=(self),
			) as pool:
				sampler = sampler_class(
					pool.loglike,
					pool.prior_transform,
					ndim=len(self.names),
					bound="multi",
					sample="auto",
					rstate=rstate,
					pool=pool,
					nlive=nlive,
				)
				sampler.run_nested(maxiter=maxiter, maxcall=maxcall, **method_kwargs)
		else:
			sampler = sampler_class(
				target_function,
				prior_transform,
				ndim=len(self.names),
				bound="multi",
				sample="auto",
				rstate=rstate,
				pool=None,
				nlive=nlive,
				logl_args=[self, bayesian_calibration_kwargs],
				ptform_args=[self],
			)
			sampler.run_nested(maxiter=maxiter, maxcall=maxcall, **method_kwargs)

		self.sampler = sampler

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		results = self.sampler.results
		for plot_func in [
			dyplot.traceplot,
			dyplot.cornerplot,
			dyplot.cornerpoints,
			dyplot.runplot,
		]:
			plot_name = plot_func.__name__.replace("_", "-")
			if plot_name == "runplot":
				plot_func(results)
			else:
				plot_func(results, labels=self.names)

			if outdir is not None:
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-{plot_name}.png",
				)
				self.append_artifact(outfile)
				plt.tight_layout()
				plt.savefig(outfile)
				plt.close()
			else:
				plt.show()
				plt.close()

		samples, weights = results.samples, results.importance_weights()
		samples_equal = dyfunc.resample_equal(samples, weights)
		trace_df = pd.DataFrame(samples_equal, columns=self.names)

		for name in trace_df:
			estimate = trace_df[name].mean()
			uncertainty = trace_df[name].std()

			parameter_estimate = ParameterEstimateModel(
				name=name, estimate=estimate, uncertainty=uncertainty
			)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(trace_df, "trace")



"""Contains the implementations for optimisation methods using Optuna

Implements the supported optimisation methods using the Optuna library.

"""

from collections.abc import Callable

import numpy as np
import optuna
import optuna.samplers as opt_samplers
import pandas as pd

from ..base import CalibrationWorkflowBase
from ..data_model import DistributionModel, ParameterDataType, ParameterEstimateModel


class OptunaOptimisation(CalibrationWorkflowBase):
	"""The Optuna optimisation method class."""

	def specify(self) -> None:
		"""Specify the parameters of the model calibration procedure."""
		sampler_name = self.specification.method
		supported_samplers = dict(
			tpes=opt_samplers.TPESampler,
			cmaes=opt_samplers.CmaEsSampler,
			nsga=opt_samplers.NSGAIISampler,
			qmc=opt_samplers.QMCSampler,
			gp=opt_samplers.GPSampler,
		)
		sampler_class = supported_samplers.get(sampler_name, None)
		if sampler_class is None:
			raise ValueError(
				f"Unsupported Optuna sampler: {sampler_name}.",
				f"Supported Optuna samplers are {', '.join(supported_samplers)}",
			)
		sampler_kwargs = self.specification.method_kwargs
		if sampler_kwargs is None:
			sampler_kwargs = {}
		self.sampler = sampler_class(**sampler_kwargs)

		self.study = optuna.create_study(
			sampler=self.sampler,
			study_name=self.specification.experiment_name,
			directions=self.specification.directions,
			storage=self.specification.storage,
			load_if_exists=True,
		)

	def execute(self) -> None:
		"""Execute the simulation calibration procedure."""
		parameter_spec = self.specification.parameter_spec.parameters

		def objective(
			trial: optuna.trial.Trial,
			parameter_spec: list[DistributionModel],
			observed_data: np.ndarray | pd.DataFrame,
			call_calibration_func: Callable,
			objective_kwargs: dict,
		) -> float | list[float]:
			parameters = {}
			for spec in parameter_spec:
				parameter_name = spec.name
				lower_bound, upper_bound = self.get_parameter_bounds(spec)
				data_type = spec.data_type

				if data_type == ParameterDataType.CONTINUOUS:
					parameters[parameter_name] = trial.suggest_float(
						parameter_name, lower_bound, upper_bound
					)
				else:
					parameters[parameter_name] = trial.suggest_int(
						parameter_name, lower_bound, upper_bound
					)

			simulation_id = self.get_simulation_uuid()
			return call_calibration_func(
				parameters, simulation_id, observed_data, **objective_kwargs
			)

		objective_kwargs = self.get_calibration_func_kwargs()

		self.study.optimize(
			lambda trial: objective(
				trial,
				parameter_spec,  # type: ignore[arg-type]
				self.specification.observed_data,
				call_calibration_func=self.call_calibration_func,
				objective_kwargs=objective_kwargs,
			),
			n_trials=self.specification.n_iterations,
			n_jobs=self.specification.n_jobs,
		)

	def analyze(self) -> None:
		"""Analyze the results of the simulation calibration procedure."""
		task, time_now, experiment_name, outdir = self.prepare_analyze()

		directions = self.specification.directions
		output_labels = self.specification.output_labels
		if output_labels is None:
			output_labels = [f"objective_{i+1}" for i in range(len(directions))]

		for i in range(len(output_labels)):
			for plot_func in [
				optuna.visualization.plot_edf,
				optuna.visualization.plot_optimization_history,
				optuna.visualization.plot_parallel_coordinate,
				optuna.visualization.plot_param_importances,
				optuna.visualization.plot_slice,
			]:
				output_label = output_labels[i]
				optimisation_plot = plot_func(
					self.study,
					target=lambda t: t.values[i],
					target_name=output_label,
				)
				plot_name = plot_func.__name__.replace("_", "-")
				if outdir is not None:
					outfile = self.join(
						outdir,
						f"{time_now}-{task}-{experiment_name}-{plot_name}-{output_label}.png",
					)
					self.append_artifact(outfile)
					optimisation_plot.write_image(outfile)
				else:
					optimisation_plot.show()

		if len(output_labels) == 2:
			optimisation_plot = optuna.visualization.plot_pareto_front(
				self.study, target_names=output_labels
			)
			if outdir is not None:
				outfile = self.join(
					outdir,
					f"{time_now}-{task}-{experiment_name}-plot-pareto-front.png",
				)
				self.append_artifact(outfile)
				optimisation_plot.write_image(outfile)
			else:
				optimisation_plot.show()

		trials_df: pd.DataFrame = self.study.trials_dataframe()
		values = [value for value in trials_df.columns if value.startswith("value")]

		trials_df_best = trials_df.sort_values(values).head(1)
		for col in trials_df_best.columns:
			if not col.startswith("params_"):
				continue
			name = col.replace("params_", "")
			estimate = trials_df_best[col].item()
			parameter_estimate = ParameterEstimateModel(name=name, estimate=estimate)
			self.add_parameter_estimate(parameter_estimate)

		if outdir is None:
			return

		self.to_csv(trials_df, "trials")
